@@ -1,23 +1,18 @@
-from __future__ import division
-from __future__ import print_function
-from __future__ import absolute_import
-from __future__ import unicode_literals
-
 import uuid
-import urllib2
-import urlparse
+import traceback
 import random
+from urllib.request import urlopen, URLError
+from urllib.parse import parse_qs
+from fsgs.context import fsgs
 from .IRC import IRC
 from .Channel import Channel
 from .IRCColor import IRCColor
 from .IRCBroadcaster import IRCBroadcaster
 from .ConnectionTester import ConnectionTester
-from ..Amiga import Amiga
+from fsgs.amiga.Amiga import Amiga
 from ..Config import Config
-from ..Database import Database
-from ..builtin_configs import find_downloadable_file
 from ..server.Server import Server
-from ..server.ServerWindow import ServerWindow
+
 
 class Netplay:
 
@@ -31,7 +26,7 @@ class Netplay:
 
     @classmethod
     def new_host_port(cls):
-        if not cls.host_ports is None:
+        if cls.host_ports is not None:
             cls.host_ports = list(range(25102, 25500))
             random.shuffle(cls.host_ports)
         return cls.host_ports.pop()
@@ -51,8 +46,8 @@ class Netplay:
         try:
             del cls.players[nick]
         except Exception:
-            print("warning, tried to remove player (but wasn't in list):",
-                    nick)
+            print(
+                "warning, tried to remove player (but wasn't in list):", nick)
         else:
             print("players are now", cls.players)
 
@@ -63,8 +58,8 @@ class Netplay:
                 cls.game_info("you are marked as ready, initiating game start")
                 Netplay.initiate_start_sequence()
             else:
-                cls.game_info("you are marked as ready "
-                        "(only ops can start game)")
+                cls.game_info(
+                    "you are marked as ready (only ops can start game)")
         else:
             cls.game_info("you are not ready")
 
@@ -78,8 +73,8 @@ class Netplay:
         if not cls.game_channel:
             cls.game_info("cannot start - no game channel")
         cls.start_sequence = str(uuid.uuid4())
-        message = u"__prestart {0} {1}".format(cls.start_sequence,
-                cls.get_config_hash())
+        message = "__prestart {0} {1}".format(
+            cls.start_sequence, cls.get_config_hash())
         IRC.channel(cls.game_channel).privmsg(message)
         # in case one wants to try a single-player game for testing
         cls.on_ackstart()
@@ -100,12 +95,12 @@ class Netplay:
                 return
         want_players = int(Config.get("__netplay_players") or "0")
         if player_count != want_players:
-            cls.game_warning(u"cannot start game, wanted {0} players, "
-                    "has {1}".format(want_players, player_count))
+            cls.game_warning("cannot start game, wanted {0} players, "
+                             "has {1}".format(want_players, player_count))
             return
         cls.game_info("sending game start command to all clients!")
-        message = u"__start {0} {1}".format(cls.start_sequence,
-                cls.get_config_hash())
+        message = "__start {0} {1}".format(
+            cls.start_sequence, cls.get_config_hash())
         IRC.channel(cls.game_channel).privmsg(message)
         cls.do_start_game()
 
@@ -117,8 +112,8 @@ class Netplay:
         # we now reset __netplay_addresses so that the ConnectionTester
         # does not connect to the game in progress
         Config.set("__netplay_addresses", "")
-        from ..LaunchHandler import LaunchHandler
-        LaunchHandler.start_local_game()
+        from ..FSUAELauncher import FSUAELauncher
+        FSUAELauncher.start_local_game()
 
         # game done, resetting config as the same game server cannot be
         # used again
@@ -138,11 +133,11 @@ class Netplay:
         if not Config.get("__netplay_addresses"):
             cls.game_warning("cannot set ready - no server configured")
             cls.game_info("an operator needs to use /startgame or "
-                    "/hostgame (or /customgame) first")
+                          "/hostgame (or /customgame) first")
             return False
         if not Config.get("__netplay_host"):
             cls.game_warning("cannot set ready - no connection to game "
-                    "server")
+                             "server")
             return False
         Config.set("__netplay_ready", "1")
         # check in case an component has force __netplay_ready back to 0
@@ -161,25 +156,25 @@ class Netplay:
 
     @classmethod
     def on_config(cls, key, value):
-        if key == "__netplay_host_last_error":
-            cls.game_warning(u"problem connecting to game server: " + value)
+        if key == "__netplay_host_last_error" and value:
+            cls.game_warning("problem connecting to game server: " + value)
         elif key == "__netplay_host":
             if value:
-                cls.game_info(u"successfully connected to net play host: "
-                        + value)
+                cls.game_info("successfully connected to net play host: " +
+                              value)
         if not IRC.running:
             return
         if not cls.game_channel:
             return
         if key in Config.sync_keys_set:
-            message = u"__config {0} {1}".format(key, value)
+            message = "__config {0} {1}".format(key, value)
             IRC.channel(cls.game_channel).privmsg(message)
 
     @classmethod
     def game_message(cls, message, color=None):
         if not cls.game_channel:
-            IRC.warning(u"game message received (but not "
-                    u"currently in a game channel): " + message)
+            IRC.warning("game message received (but not "
+                        "currently in a game channel): " + message)
             return
         IRC.channel(cls.game_channel).message(message, color)
         if cls.game_channel != IRC.active_channel:
@@ -228,7 +223,8 @@ class Netplay:
                 if args["nick"] != IRC.my_nick:
                     if cls.is_op():
                         # operator - send config to new player(s)
-                        IRC.channel(cls.game_channel).action(
+                        IRC.channel(
+                            cls.game_channel).action(
                                 "sends config on arrival of new player")
                         cls.send_config()
         elif key == "part":
@@ -239,7 +235,7 @@ class Netplay:
     def on_join_game_channel(cls, channel):
         # first we erase config which is not part of the default configuration
         # key set, to avoid some problems
-        keys = Config.config.keys()
+        keys = fsgs.config.values.keys()
         for key in keys:
             if key not in Config.default_config:
                 Config.set(key, "")
@@ -253,12 +249,13 @@ class Netplay:
 
     @classmethod
     def reset_netplay_config(cls):
-        Config.set_multiple([("__netplay_id", ""),
-                ("__netplay_password", ""),
-                ("__netplay_players", ""),
-                ("__netplay_port", ""),
-                ("__netplay_addresses", ""),
-                ("__netplay_host", "")])
+        Config.set_multiple([
+            ("__netplay_id", ""),
+            ("__netplay_password", ""),
+            ("__netplay_players", ""),
+            ("__netplay_port", ""),
+            ("__netplay_addresses", ""),
+            ("__netplay_host", "")])
 
     @classmethod
     def handle_command(cls, command):
@@ -266,7 +263,6 @@ class Netplay:
             return False
         args = command[1:].split(" ")
         print(args)
-        command = args[0]
         function_name = "command_" + args[0]
         try:
             function = getattr(cls, function_name)
@@ -275,18 +271,24 @@ class Netplay:
         else:
             try:
                 function(args[1:])
-            except Exception, e:
-                cls.game_warning(repr(e))
+            except Exception:
+                cls.game_warning(traceback.format_exc())
             return True
 
     @classmethod
     def command_ready(cls, args):
+        if len(args) != 0:
+            IRC.warning("syntax: /ready")
+            return
         cls.set_ready()
-        if Config.get("__netplay_ready") ==  "1":
+        if Config.get("__netplay_ready") == "1":
             cls.game_info("status set to ready")
 
     @classmethod
     def command_notready(cls, args):
+        if len(args) != 0:
+            IRC.warning("syntax: /notready")
+            return
         Config.set("__netplay_ready", "0")
         cls.game_info("status set to not ready")
 
@@ -303,7 +305,7 @@ class Netplay:
             return
         for arg in args:
             cls.print_check_request(IRC.my_nick, arg, Config.get(arg))
-            message = u"__check {0} {1}".format(arg, Config.get(arg))
+            message = "__check {0} {1}".format(arg, Config.get(arg))
             IRC.channel(cls.game_channel).privmsg(message)
 
     @classmethod
@@ -317,16 +319,16 @@ class Netplay:
         if not channel.is_op():
             IRC.warning("verify: you need to be an operator")
             return
-        #for arg in args:
-        #    cls.game_info("requesting verification of {0} = {1}".format(arg,
-        #            Config.get(arg)))
+        # for arg in args:
+        #     cls.game_info("requesting verification of {0} = {1}".format(arg,
+        #             Config.get(arg)))
         cls.print_verify_request(IRC.my_nick)
-        IRC.channel(cls.game_channel).privmsg(u"__beginverify")
+        IRC.channel(cls.game_channel).privmsg("__beginverify")
         for key in Config.checksum_keys:
             value = Config.get(key)
-            message = u"__verify {0} {1}".format(key, value)
+            message = "__verify {0} {1}".format(key, value)
             IRC.channel(cls.game_channel).privmsg(message)
-        IRC.channel(cls.game_channel).privmsg(u"__endverify")
+        IRC.channel(cls.game_channel).privmsg("__endverify")
 
     @classmethod
     def command_sendconfig(cls, args):
@@ -347,7 +349,7 @@ class Netplay:
         # sending keys in preferred key order
         for key in Config.sync_keys_list:
             value = Config.get(key)
-            message = u"__config {0} {1}".format(key, value)
+            message = "__config {0} {1}".format(key, value)
             channel.privmsg(message)
 
     @classmethod
@@ -361,6 +363,9 @@ class Netplay:
 
     @classmethod
     def command_resetconfig(cls, args):
+        if len(args) != 0:
+            IRC.warning("syntax: /resetconfig")
+            return
         channel = Channel.active()
         if not cls.require_game_channel(channel):
             return
@@ -408,7 +413,7 @@ class Netplay:
         result = cls.parse_server_args(args, 25101)
         if not result:
             IRC.warning("usage: /startgame <host>[:<port>] "
-                    "[<players>] [<password>]")
+                        "[<players>] [<password>]")
             return
         host, port, players, password = result
 
@@ -416,30 +421,35 @@ class Netplay:
             IRC.warning("startgame: you need to be an operator")
             return
 
-        url = u"http://{0}:{1}/game/create?players={2}&password={3}".format(
-                host, port, players, password)
+        url = "http://{0}:{1}/game/create?players={2}&password={3}".format(
+            host, port, players, password)
         try:
-            f = urllib2.urlopen(url)
+            f = urlopen(url)
             result = f.read()
-        except urllib2.URLError, e:
-            channel.warning(u"problem starting game server: {0}".format(
-                    repr(e)))
+        except URLError:
+            channel.warning(
+                "Problem starting game server: {0}".format(
+                    traceback.format_exc()))
         else:
+            result = result.decode("UTF-8")
             print(result)
-            result_dict = urlparse.parse_qs(result)
+            result_dict = parse_qs(result)
+            print(result_dict)
             game_id = result_dict["id"][0]
             game_password = result_dict["password"][0]
             game_port = result_dict["port"][0]
             game_addresses = result_dict["addresses"][0]
-            Config.set_multiple([("__netplay_game", game_id),
-                    ("__netplay_password", game_password),
-                    ("__netplay_players", str(players)),
-                    ("__netplay_port", game_port),
-                    ("__netplay_host", u""),
-                    ("__netplay_addresses", game_addresses)])
-            channel.info(u"started game id: {0} password: {1} "
-                    "server: {2} port: {3}".format(
-                    game_id, game_password, game_addresses, game_port))
+            Config.set_multiple([
+                ("__netplay_game", game_id),
+                ("__netplay_password", game_password),
+                ("__netplay_players", str(players)),
+                ("__netplay_port", game_port),
+                ("__netplay_host", ""),
+                ("__netplay_addresses", game_addresses)])
+            channel.info("started game id: {0} password: {1} "
+                         "server: {2} port: {3}".format(
+                             game_id, game_password, game_addresses,
+                             game_port))
 
     @classmethod
     def command_hostgame(cls, args):
@@ -448,10 +458,10 @@ class Netplay:
             return
         result = cls.parse_server_args(args, cls.new_host_port())
         if not result:
-            IRC.warning("usage: /hostgame <host>[:<port>] "
-                    "[<players>] [<password>]")
+            IRC.warning("usage: /hostgame <address>[,<address2>][:<port>] "
+                        "[<players>] [<password>]")
             return
-        host, port, players, password = result
+        addresses, port, players, password = result
 
         if not channel.is_op():
             IRC.warning("hostgame: you need to be an operator")
@@ -459,19 +469,22 @@ class Netplay:
 
         server = Server(port, players, password)
         server.start()
+
+        from ..server.ServerWindow import ServerWindow
         window = ServerWindow(None, server)
-        window.Show()
+        window.show()
 
         game_id = str(uuid.uuid4())
-        Config.set_multiple([("__netplay_game", game_id),
-                ("__netplay_password", password),
-                ("__netplay_players", str(players)),
-                ("__netplay_port", str(port)),
-                ("__netplay_host", u""),
-                ("__netplay_addresses", host)])
-        channel.info(u"started game id: {0} password: {1} "
-                "server: {2} port: {3}".format(
-                game_id, password, host, port))
+        Config.set_multiple([
+            ("__netplay_game", game_id),
+            ("__netplay_password", password),
+            ("__netplay_players", str(players)),
+            ("__netplay_port", str(port)),
+            ("__netplay_host", ""),
+            ("__netplay_addresses", addresses)])
+        channel.info("started game id: {0} password: {1} "
+                     "server: {2} port: {3}".format(
+                         game_id, password, addresses, port))
 
     @classmethod
     def command_setserver(cls, args):
@@ -484,8 +497,8 @@ class Netplay:
             return
         result = cls.parse_server_args(args, 25100)
         if not result:
-            IRC.warning("usage: /customgame <addresses>[:<port>] "
-                    "[<players>] [<password>]")
+            IRC.warning("usage: /customgame <address>[,<address2>][:<port>] "
+                        "[<players>] [<password>]")
             return
         addresses, port, players, password = result
 
@@ -494,15 +507,16 @@ class Netplay:
             return
 
         game_id = str(uuid.uuid4())
-        Config.set_multiple([("__netplay_game", game_id),
-                ("__netplay_password", password),
-                ("__netplay_players", str(players)),
-                ("__netplay_port", str(port)),
-                ("__netplay_host", u""),
-                ("__netplay_addresses", addresses)])
-        channel.info(u"started game id: {0} password: {1} "
-                "server: {2} port: {3}".format(
-                game_id, password, addresses, port))
+        Config.set_multiple([
+            ("__netplay_game", game_id),
+            ("__netplay_password", password),
+            ("__netplay_players", str(players)),
+            ("__netplay_port", str(port)),
+            ("__netplay_host", ""),
+            ("__netplay_addresses", addresses)])
+        channel.info("started game id: {0} password: {1} "
+                     "server: {2} port: {3}".format(
+                         game_id, password, addresses, port))
 
     @classmethod
     def command_set(cls, args):
@@ -526,16 +540,16 @@ class Netplay:
     def handle_game_instruction(cls, nick, message):
         channel = IRC.channel(cls.game_channel)
 
-        #if Netplay.game_channel != self.name:
-        #    self.warning(self, u"ignored command in "
-        #            "non-active game: {0}".format(command), IRCColor.WARNING)
-        #    return
-        words = message.split(u" ")
+        # if Netplay.game_channel != self.name:
+        #     self.warning(self, "ignored command in "
+        #             "non-active game: {0}".format(command), IRCColor.WARNING)
+        #     return
+        words = message.split(" ")
         command = words[0]
-        arg = u" ".join(words[1:])
+        arg = " ".join(words[1:])
         if command == "__config":
             if channel.is_op(nick):
-                args = arg.split(u" ", 1)
+                args = arg.split(" ", 1)
                 key = args[0]
                 value = ""
                 if len(args) > 1:
@@ -543,30 +557,30 @@ class Netplay:
                 cls.set_config(key.strip(), value.strip())
         elif command == "__prestart":
             if channel.is_op(nick):
-                seq, config_hash = arg.split(u" ")
+                seq, config_hash = arg.split(" ")
                 my_config_hash = cls.get_config_hash()
                 if my_config_hash != config_hash:
                     channel.privmsg("not the same config hash")
                     return
                 if Config.get("__netplay_ready") != "1":
                     channel.privmsg("(op tried to start game, "
-                            "but I am not ready)")
+                                    "but I am not ready)")
                     return
-                channel.privmsg("__ackstart {0} {1}".format(seq,
-                    my_config_hash))
+                channel.privmsg("__ackstart {0} {1}".format(
+                    seq, my_config_hash))
         elif command == "__ackstart":
-            start_sequence, config_hash = arg.split(u" ")
+            start_sequence, config_hash = arg.split(" ")
             cls.player(nick).set("start_sequence", start_sequence)
             cls.player(nick).set("config_hash", config_hash)
             cls.on_ackstart()
             # FIXME: check start sequence?
         elif command == "__start":
             if channel.is_op(nick):
-                start_sequence, config_hash = arg.split(u" ")
+                start_sequence, config_hash = arg.split(" ")
                 my_config_hash = Netplay.get_config_hash()
                 if my_config_hash != config_hash:
                     channel.action("could not start game "
-                            "(mismatching config hash)")
+                                   "(mismatching config hash)")
                     return
                 Netplay.do_start_game()
         elif command == "__resetconfig":
@@ -575,46 +589,47 @@ class Netplay:
                 Netplay.reset_config()
         elif command == "__check":
             if channel.is_op(nick):
-                args = arg.split(u" ")
+                args = arg.split(" ")
                 key = args[0]
-                value = u" ".join(args[1:])
+                value = " ".join(args[1:])
                 my_value = Config.get(key)
                 cls.print_check_request(nick, key, value)
                 cls.print_check_response(IRC.my_nick, key, my_value, value)
-                channel.privmsg(u"__ackcheck {0} {1}".format(key, my_value))
+                channel.privmsg("__ackcheck {0} {1}".format(key, my_value))
         elif command == "__beginverify":
             if channel.is_op(nick):
                 cls.print_verify_request(nick)
         elif command == "__verify":
             if channel.is_op(nick):
-                args = arg.split(u" ")
+                args = arg.split(" ")
                 key = args[0]
-                value = u" ".join(args[1:])
+                value = " ".join(args[1:])
                 if Config.get(key) != value:
-                    cls.print_verify_response(IRC.my_nick, key,
-                            Config.get(key))
-                    channel.privmsg(u"__ackverify {0} {1}".format(key,
-                        Config.get(key)))
+                    cls.print_verify_response(
+                        IRC.my_nick, key, Config.get(key))
+                    channel.privmsg("__ackverify {0} {1}".format(
+                        key, Config.get(key)))
         elif command == "__endverify":
             pass
         elif command == "__ackcheck":
-            args = arg.split(u" ")
+            args = arg.split(" ")
             key = args[0]
-            value = u" ".join(args[1:])
+            value = " ".join(args[1:])
             my_value = Config.get(key)
             cls.print_check_response(nick, key, value, my_value)
         elif command == "__ackverify":
-            args = arg.split(u" ", 1)
-            cls.print_verify_response(nick, args[0], u" ".join(args[1:]))
+            args = arg.split(" ", 1)
+            cls.print_verify_response(nick, args[0], " ".join(args[1:]))
         else:
-            channel.warning(u"unknown command received: {0}".format(
-                    command))
+            channel.warning(
+                "unknown command received: {0}".format(command))
 
     @classmethod
     def print_check_request(cls, nick, key, value):
-        IRC.channel(cls.game_channel).info(
-                u"* {0} requested check of {1} ({2})".format(
-                nick, key, value))
+        IRC.channel(
+            cls.game_channel).info(
+                "* {0} requested check of {1} ({2})".format(
+                    nick, key, value))
 
     @classmethod
     def print_check_response(cls, nick, key, value, check_value):
@@ -622,24 +637,28 @@ class Netplay:
             color = IRCColor.INFO
         else:
             color = IRCColor.WARNING
-        IRC.channel(cls.game_channel).message(
-                u"{0} = {1} ({2})".format(key, value, nick), color)
+        IRC.channel(
+            cls.game_channel).message(
+                "{0} = {1} ({2})".format(key, value, nick), color)
 
     @classmethod
     def print_verify_request(cls, nick):
-        IRC.channel(cls.game_channel).info(
-                u"* {0} requested automatic config verification".format(nick))
+        IRC.channel(
+            cls.game_channel).info(
+                "* {0} requested automatic config verification".format(nick))
 
     @classmethod
     def print_verify_response(cls, nick, key, value):
-        IRC.channel(cls.game_channel).warning(
-                u"{0} = {1} ({2})".format(key, value, nick))
+        IRC.channel(
+            cls.game_channel).warning(
+                "{0} = {1} ({2})".format(key, value, nick))
 
     @classmethod
     def set_config(cls, key, value):
         # this config was received from a game channel operator
         print("received config", key, value)
-        channel = IRC.channel(cls.game_channel)
+        # channel =
+        IRC.channel(cls.game_channel)
         if key not in Config.sync_keys_set:
             print("not processing this key")
             return
@@ -656,35 +675,35 @@ class Netplay:
         if key == "x_kickstart_file_sha1":
             if value == Amiga.INTERNAL_ROM_SHA1:
                     Config.set_multiple([
-                            ("kickstart_file", "internal"),
-                            ("x_kickstart_file", "internal"),
-                            (key, value)])
+                        ("kickstart_file", "internal"),
+                        ("x_kickstart_file", "internal"),
+                        (key, value)])
             else:
-                path = Database.get_instance().find_file(sha1=value)
+                path = fsgs.file.find_by_sha1(sha1=value)
                 if path:
                     Config.set_multiple([
-                            ("kickstart_file", path),
-                            ("x_kickstart_file", path),
-                            (key, value)])
+                        ("kickstart_file", path),
+                        ("x_kickstart_file", path),
+                        (key, value)])
                 else:
-                    channel.action(u"could not find kickstart for {0}".format(
-                            repr(value)))
+                    channel.action("could not find kickstart for {0}".format(
+                                   repr(value)))
         elif key == "x_kickstart_ext_file_sha1":
             if not value:
                 Config.set_multiple([
-                        ("kickstart_ext_file", ""),
-                        ("x_kickstart_ext_file", ""),
-                        (key, "")])
+                    ("kickstart_ext_file", ""),
+                    ("x_kickstart_ext_file", ""),
+                    (key, "")])
                 return
-            path = Database.get_instance().find_file(sha1=value)
+            path = fsgs.file.find_by_sha1(sha1=value)
             if path:
                 Config.set_multiple([
-                        ("kickstart_ext_file", path),
-                        ("x_kickstart_ext_file", path),
-                        (key, value)])
+                    ("kickstart_ext_file", path),
+                    ("x_kickstart_ext_file", path),
+                    (key, value)])
             else:
-                channel.action(u"could not find (ext) kickstart "
-                        "for {0}".format(repr(value)))
+                channel.action("could not find (ext) kickstart "
+                               "for {0}".format(repr(value)))
 
     @classmethod
     def set_file_config(cls, key, value):
@@ -696,33 +715,35 @@ class Netplay:
         if value.startswith("http://") or value.startswith("https://"):
             path = value
         else:
-            path = Database.get_instance().find_file(sha1=value)
-        if not path:
-            path = find_downloadable_file(value)
+            # path = Database.get_instance().find_file(sha1=value)
+            path = fsgs.file.find_by_sha1(sha1=value)
+        # if not path:
+        #     path = find_downloadable_file(value)
         if path:
             Config.set_multiple([(set_key, path), (key, value)])
         else:
             Config.set_multiple([(set_key, ""), (key, "")])
-            channel.action(u"could not find {1} for "
-                    "for {0}".format(value, set_key))
+            channel.action("could not find {1} for "
+                           "for {0}".format(value, set_key))
 
     file_config = {
     }
     for i in range(Amiga.MAX_FLOPPY_DRIVES):
         file_config["x_floppy_drive_{0}_sha1".format(i)] = \
-                "floppy_drive_{0}".format(i)
+            "floppy_drive_{0}".format(i)
     for i in range(Amiga.MAX_FLOPPY_IMAGES):
         file_config["x_floppy_image_{0}_sha1".format(i)] = \
-                "floppy_image_{0}".format(i)
+            "floppy_image_{0}".format(i)
     for i in range(Amiga.MAX_CDROM_DRIVES):
         file_config["x_cdrom_drive_{0}_sha1".format(i)] = \
-                "cdrom_drive_{0}".format(i)
+            "cdrom_drive_{0}".format(i)
     for i in range(Amiga.MAX_CDROM_IMAGES):
         file_config["x_cdrom_image_{0}_sha1".format(i)] = \
-                "cdrom_image_{0}".format(i)
+            "cdrom_image_{0}".format(i)
     for i in range(Amiga.MAX_HARD_DRIVES):
         file_config["x_hard_drive_{0}_sha1".format(i)] = \
-                "hard_drive_{0}".format(i)
+            "hard_drive_{0}".format(i)
+
 
 Config.add_listener(Netplay)
 IRCBroadcaster.add_listener(Netplay)
