@@ -1,9 +1,5 @@
-from __future__ import division
-from __future__ import print_function
-from __future__ import absolute_import
-from __future__ import unicode_literals
+import fsui as fsui
 
-import fs_uae_launcher.fsui as fsui
 
 class Book(fsui.Panel):
 
