@@ -1,15 +1,9 @@
-from __future__ import division
-from __future__ import print_function
-from __future__ import absolute_import
-from __future__ import unicode_literals
-
-import fs_uae_launcher.fsui as fsui
-from ..Config import Config
-from ..I18N import _, ngettext
+import fsui as fsui
 from .FloppiesGroup import FloppiesGroup
 from .MediaListGroup import MediaListGroup
 from .Skin import Skin
 
+
 class FloppiesPanel(fsui.Panel):
 
     def __init__(self, parent):
