@@ -1,13 +1,7 @@
-from __future__ import division
-from __future__ import print_function
-from __future__ import absolute_import
-from __future__ import unicode_literals
-
-import fs_uae_launcher.fsui as fsui
-from ..Config import Config
-from ..I18N import _, ngettext
+import fsui as fsui
 from .Skin import Skin
 
+
 class InputPanel(fsui.Panel):
 
     def __init__(self, parent):
@@ -22,9 +16,9 @@ class InputPanel(fsui.Panel):
         self.layout.add_spacer(Skin.EXTRA_GROUP_MARGIN)
 
         self.input_group = InputGroup(self, with_more_options=False,
-                parallel_ports=True)
+                                      parallel_ports=True)
         self.layout.add(self.input_group, fill=True)
 
         self.input_group = InputGroup(self, with_more_options=False,
-                custom_ports=True)
+                                      custom_ports=True)
         self.layout.add(self.input_group, fill=True)
