@@ -1,19 +1,10 @@
-from __future__ import division
-from __future__ import print_function
-from __future__ import absolute_import
-from __future__ import unicode_literals
-
-import os
-import fs_uae_launcher.fsui as fsui
-from ..Amiga import Amiga
+from fsbc.util import unused
+import fsui as fsui
+from fsgs.amiga.Amiga import Amiga
 from ..Config import Config
-from ..CDManager import CDManager
-from ..FloppyManager import FloppyManager
-from ..I18N import _, ngettext
-from ..Settings import Settings
-from .IconButton import IconButton
 from .FloppiesGroup import FloppiesGroup
 
+
 class RemovableMediaGroup(FloppiesGroup):
 
     def __init__(self, parent, drives):
@@ -32,11 +23,12 @@ class RemovableMediaGroup(FloppiesGroup):
         Config.remove_listener(self)
 
     def on_config(self, key, value):
+        unused(value)
         if key == "amiga_model":
             self.update_media_type()
 
     def update_media_type(self):
-        self.set_cd_mode(Amiga.is_cd_based())
+        self.set_cd_mode(Amiga.is_cd_based(Config))
 
     def set_cd_mode(self, cd_mode):
         if self.cd_mode == cd_mode:
