@@ -1,11 +1,7 @@
-from __future__ import division
-from __future__ import print_function
-from __future__ import absolute_import
-from __future__ import unicode_literals
-
-import fs_uae_launcher.fsui as fsui
+import fsui as fsui
 from ...Config import Config
 
+
 class ConfigCheckBox(fsui.CheckBox):
 
     def __init__(self, parent, label, config_key):
