@@ -1,5 +1,5 @@
-#ifndef FS_THREAD_H_
-#define FS_THREAD_H_
+#ifndef FS_THREAD_H
+#define FS_THREAD_H
 
 #include <stdint.h>
 
@@ -13,9 +13,15 @@ struct fs_thread;
 typedef struct fs_thread fs_thread;
 
 typedef void *(*fs_thread_function)(void *);
-fs_thread *fs_thread_create(fs_thread_function fn, void *data);
+fs_thread *fs_thread_create(
+        const char *name, fs_thread_function fn, void *data);
+#if 0
+fs_thread *fs_thread_create_detached(
+        const char *name, fs_thread_function fn, void *data);
+#endif
 void *fs_thread_wait(fs_thread *thread);
-void fs_thread_destroy(fs_thread *thread);
+//void fs_thread_destroy(fs_thread *thread);
+void fs_thread_free(fs_thread *thread);
 
 struct fs_mutex;
 typedef struct fs_mutex fs_mutex;
@@ -32,7 +38,8 @@ fs_condition *fs_condition_create(void);
 void fs_condition_destroy(fs_condition *condition);
 int fs_condition_signal(fs_condition *condition);
 int fs_condition_wait(fs_condition *condition, fs_mutex *mutex);
-int fs_condition_timed_wait(fs_condition *condition, fs_mutex *mutex,
+int64_t fs_condition_get_wait_end_time(int period);
+int fs_condition_wait_until(fs_condition *condition, fs_mutex *mutex,
         int64_t real_time);
 
 struct fs_semaphore;
@@ -48,4 +55,4 @@ int fs_semaphore_try_wait(fs_semaphore *semaphore);
 }
 #endif
 
-#endif // FS_THREAD_H_
+#endif /* FS_THREAD_H */
