@@ -1,10 +1,15 @@
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
 #include "font.h"
 
+#include <fs/emu.h>
+#include <fs/glib.h>
+#include <fs/i18n.h>
+
 #include <stdlib.h>
 #include <string.h>
-#include <fs/i18n.h>
-#include <fs/list.h>
-#include <fs/string.h>
 
 #ifdef USE_OPENGL
 #include <fs/ml/opengl.h>
@@ -27,7 +32,7 @@ static int g_texture_height = 2048;
 #define MASK 0x00ffffff
 
 static int g_initialized = 0;
-static fs_list* g_cache = NULL;
+static GList* g_cache = NULL;
 static int g_video_version = 0;
 static GLuint g_text_texture = 0;
 static uint8_t *g_buffer = NULL;
@@ -52,13 +57,13 @@ typedef struct _cache_item {
 static void sanity_check() {
 }
 
-void initialize_cache() {
+static void initialize_cache() {
     for (int i = 0; i < CACHE_SIZE; i++) {
         cache_item *item = malloc(sizeof(cache_item));
         item->font = NULL;
         item->text = NULL;
         item->position = i;
-        g_cache = fs_list_append(g_cache, item);
+        g_cache = g_list_append(g_cache, item);
     }
     sanity_check();
     // FIXME: REMOVE
@@ -70,10 +75,10 @@ static void create_text_texture() {
     fs_gl_bind_texture(g_text_texture);
     // want to clear data to color (0, 0, 0, 0), probably a better
     // way to to this...
-    void *data = fs_malloc0(g_texture_width * g_texture_height * 4);
+    void *data = g_malloc0(g_texture_width * g_texture_height * 4);
     glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, g_texture_width, g_texture_height,
             0, GL_RGBA, GL_UNSIGNED_BYTE, data);
-    free(data);
+    g_free(data);
     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
@@ -90,14 +95,14 @@ static void context_notification_handler(int notification, void *data) {
         // FIXME: clear text cache..
         //printf("FIXME: clear text cache\n");
 
-        fs_list* list = g_cache;
+        GList* list = g_cache;
         while (list) {
             cache_item *item = (cache_item *) list->data;
             free(item->text);
             free(item);
             list = list->next;
         }
-        fs_list_free(g_cache);
+        g_list_free(g_cache);
         g_cache = NULL;
         initialize_cache();
     }
@@ -108,7 +113,7 @@ static void context_notification_handler(int notification, void *data) {
 
 #ifdef USE_FREETYPE
 
-void init_freetype(void) {
+static void init_freetype(void) {
     int error = FT_Init_FreeType(&library);
     if (error) {
         fs_emu_warning("Could not initialize freetype");
@@ -120,7 +125,7 @@ void init_freetype(void) {
 
 #endif
 
-void initialize() {
+static void initialize() {
     g_texture_width = 2048;
     g_texture_height = 2048;
     int max_texture_size = fs_ml_get_max_texture_size();
@@ -148,7 +153,7 @@ void initialize() {
 
 void fs_emu_font_measure(fs_emu_font *font, const char *text, int* width,
         int *height) {
-    if (font->image == NULL) {
+    if (font == NULL || font->image == NULL) {
         if (width) {
             *width = 0;
         }
@@ -184,7 +189,7 @@ int fs_emu_font_render_with_outline(fs_emu_font *font, const char *text,
 }
 
 //gunichar2 fix_char(fs_emu_font *font, gunichar2 c) {
-int fix_char(fs_emu_font *font, int c) {
+static int fix_char(fs_emu_font *font, int c) {
     if (c == 0x2019) {
         // replace RIGHT SINGLE QUOTATION MARK with common apostrophe
         return '\'';
@@ -203,7 +208,7 @@ int fix_char(fs_emu_font *font, int c) {
 
 int fs_emu_font_render(fs_emu_font *font, const char *text, float x, float y,
         float r, float g, float b, float alpha) {
-    if (font->image == NULL) {
+    if (font == NULL || font->image == NULL) {
         return 0;
     }
     if (text == NULL || *text == '\0') {
@@ -214,7 +219,7 @@ int fs_emu_font_render(fs_emu_font *font, const char *text, float x, float y,
     }
 
     // find cached text entry, if any
-    fs_list* list = g_cache;
+    GList* list = g_cache;
     while (list) {
         cache_item *item = (cache_item *) list->data;
         if (item->font == font && strcmp(item->text, text) == 0) {
@@ -224,7 +229,7 @@ int fs_emu_font_render(fs_emu_font *font, const char *text, float x, float y,
     }
     if (list) {
         cache_item *item = (cache_item *) list->data;
-        g_cache = fs_list_delete_link(g_cache, list);
+        g_cache = g_list_delete_link(g_cache, list);
         sanity_check();
         fs_gl_blending(1);
         fs_gl_texturing(1);
@@ -272,7 +277,7 @@ int fs_emu_font_render(fs_emu_font *font, const char *text, float x, float y,
         glEnd();
 #endif
 
-        g_cache = fs_list_prepend(g_cache, item);
+        g_cache = g_list_prepend(g_cache, item);
         sanity_check();
         return item->width;
     }
@@ -458,7 +463,7 @@ int fs_emu_font_render(fs_emu_font *font, const char *text, float x, float y,
     //free(utext);
     //free(base_text);
 
-    fs_list *last = fs_list_last(g_cache);
+    GList *last = g_list_last(g_cache);
     cache_item *last_item = (cache_item *) last->data;
     int position = last_item->position;
 
@@ -483,7 +488,7 @@ int fs_emu_font_render(fs_emu_font *font, const char *text, float x, float y,
 
     cache_item *item = malloc(sizeof(cache_item));
     item->font = font;
-    item->text = fs_strdup(text);
+    item->text = g_strdup(text);
     item->width = required_width;
     item->height = required_height;
     item->position = position;
@@ -494,7 +499,7 @@ int fs_emu_font_render(fs_emu_font *font, const char *text, float x, float y,
     item->y2 = (item->position * 32 + required_height) /
             (1.0 * g_texture_height);
     //item->texture = render_texture;
-    g_cache = fs_list_prepend(g_cache, item);
+    g_cache = g_list_prepend(g_cache, item);
     sanity_check();
 
     if (last_item->text) {
@@ -507,17 +512,30 @@ int fs_emu_font_render(fs_emu_font *font, const char *text, float x, float y,
     */
     free(last_item);
 
-    g_cache = fs_list_delete_link(g_cache, last);
+    g_cache = g_list_delete_link(g_cache, last);
     sanity_check();
 
     // now the text is in the cache, so call function again
     return fs_emu_font_render(font, text, x, y, r, g, b, alpha);
 }
 
-static fs_image *load_font_from_file(const char *path) {
-    //char *full_name = g_strconcat(name, ".png", NULL);
-    //char *path = g_build_filename(fs_emu_get_share_dir(), full_name, NULL);
-    //char *path = fs_get_program_data_file(full_name);
+static void convert_to_premultiplied_alpha(fs_image *image)
+{
+    if (image->format == FS_IMAGE_FORMAT_RGBA) {
+        int num_pixels = image->width * image->height;
+        unsigned char *pixels = image->data;
+        for (int i = 0; i < num_pixels; i++) {
+            unsigned char alpha = pixels[3];
+            pixels[0] = ((int) pixels[0]) * alpha / 255;
+            pixels[1] = ((int) pixels[1]) * alpha / 255;
+            pixels[2] = ((int) pixels[2]) * alpha / 255;
+            pixels += 4;
+        }
+    }
+}
+
+static fs_image *load_font_from_file(const char *path)
+{
     if (path == NULL) {
         fs_emu_warning(_("Could not find font: %s"), path);
         return NULL;
@@ -528,34 +546,23 @@ static fs_image *load_font_from_file(const char *path) {
         fs_emu_warning(_("Error loading font: %s"), path);
         return NULL;
     }
-
-    // convert to premultiplied alpha
-    if (image->format == FS_IMAGE_FORMAT_RGBA) {
-        int num_pixels = image->width * image->height;
-        unsigned char *pixels = image->data;
-        for (int i = 0; i < num_pixels; i++) {
-            unsigned char alpha = pixels[3];
-            // should really divide by 255, but 256 is faster...
-            //pixels[0] = ((int) pixels[0]) * alpha / 256;
-            //pixels[1] = ((int) pixels[1]) * alpha / 256;
-            //pixels[2] = ((int) pixels[2]) * alpha / 256;
-            pixels[0] = ((int) pixels[0]) * alpha / 255;
-            pixels[1] = ((int) pixels[1]) * alpha / 255;
-            pixels[2] = ((int) pixels[2]) * alpha / 255;
-            //pixels[0] = (unsigned char) ((pixels[0] * alpha + 0.5) / 255.0);
-            //pixels[1] = (unsigned char) ((pixels[1] * alpha + 0.5) / 255.0);
-            //pixels[2] = (unsigned char) ((pixels[2] * alpha + 0.5) / 255.0);
-            pixels += 4;
-        }
-    }
+    convert_to_premultiplied_alpha(image);
     return image;
 }
 
-fs_emu_font *fs_emu_font_new_from_file(const char *name) {
-    fs_emu_log("load font %s\n", name);
-    fs_emu_font *font = fs_malloc0(sizeof(fs_emu_font));
-    font->image = load_font_from_file(name);
+static fs_image *load_font_from_data(char *data, int size)
+{
+    fs_image *image = fs_image_new_from_data(data, size);
+    if (image == NULL) {
+        fs_emu_warning(_("Error loading font from data"));
+        return NULL;
+    }
+    convert_to_premultiplied_alpha(image);
+    return image;
+}
 
+static void prepare_font(fs_emu_font *font)
+{
     if (font->image) {
         unsigned char *data = font->image->data;
         uint32_t *idata = (uint32_t *) data;
@@ -621,6 +628,24 @@ fs_emu_font *fs_emu_font_new_from_file(const char *name) {
         font->chars = c + 1;
         fs_emu_log("%d characters\n", font->chars);
     }
+}
 
+fs_emu_font *fs_emu_font_new_from_data(char *data, int size)
+{
+    fs_emu_log("load font from data\n");
+    fs_emu_font *font = g_malloc0(sizeof(fs_emu_font));
+    if (data != NULL) {
+        font->image = load_font_from_data(data, size);
+    }
+    prepare_font(font);
+    return font;
+}
+
+fs_emu_font *fs_emu_font_new_from_file(const char *name)
+{
+    fs_emu_log("load font %s\n", name);
+    fs_emu_font *font = g_malloc0(sizeof(fs_emu_font));
+    font->image = load_font_from_file(name);
+    prepare_font(font);
     return font;
 }
