@@ -1,17 +1,26 @@
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
+#define FS_EMU_INTERNAL
+#include <fs/emu.h>
+#include <fs/emu/options.h>
+#include <fs/emu/path.h>
+
 #include "render.h"
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
-#include <fs/config.h>
+#include <fs/conf.h>
 #include <fs/filesys.h>
 #include <fs/i18n.h>
+#include <fs/glib.h>
 #include <fs/ml.h>
-#include <fs/queue.h>
-#include <fs/string.h>
 #include <fs/thread.h>
 #include <fs/time.h>
+#include <fs/util.h>
 
 #include "libfsemu.h"
 #include "audio.h"
@@ -106,7 +115,7 @@ void fs_emu_set_overlay_state(int overlay, int state) {
     if (overlay < 0 || overlay >= FS_EMU_MAX_OVERLAYS) {
         return;
     }
-    fs_mutex_lock(g_overlay_mutex);
+    // fs_mutex_lock(g_overlay_mutex);
     g_overlay_status[overlay].state = state;
 
 #if 0
@@ -119,7 +128,7 @@ void fs_emu_set_overlay_state(int overlay, int state) {
         g_overlay_status[overlay].was_disabled = 1;
     }
 #endif
-    fs_mutex_unlock(g_overlay_mutex);
+    // fs_mutex_unlock(g_overlay_mutex);
 }
 
 static void setup_opengl() {
@@ -147,11 +156,11 @@ void fs_emu_initialize_opengl() {
     fs_gl_add_context_notification(context_notification_handler, NULL);
 }
 
-void get_buffer_format(int *gl_buffer_format, int *gl_buffer_type) {
+static void get_buffer_format(int *gl_buffer_format, int *gl_buffer_type) {
     int format = fs_emu_get_video_format();
     if (format == FS_EMU_VIDEO_FORMAT_BGRA) {
         *gl_buffer_format = GL_BGRA;
-#ifdef __BIG_ENDIAN__
+#ifdef WORDS_BIGENDIAN
         *gl_buffer_type = GL_UNSIGNED_INT_8_8_8_8_REV;
 #else
         *gl_buffer_type = GL_UNSIGNED_BYTE;
@@ -159,7 +168,7 @@ void get_buffer_format(int *gl_buffer_format, int *gl_buffer_type) {
     }
     else if (format == FS_EMU_VIDEO_FORMAT_RGBA) {
         *gl_buffer_format = GL_RGBA;
-#ifdef __BIG_ENDIAN__
+#ifdef WORDS_BIGENDIAN
         *gl_buffer_type = GL_UNSIGNED_INT_8_8_8_8_REV;
 #else
         *gl_buffer_type = GL_UNSIGNED_BYTE;
@@ -167,7 +176,7 @@ void get_buffer_format(int *gl_buffer_format, int *gl_buffer_type) {
     }
     else if (format == FS_EMU_VIDEO_FORMAT_RGB) {
         *gl_buffer_format = GL_RGB;
-#ifdef __BIG_ENDIAN__
+#ifdef WORDS_BIGENDIAN
         *gl_buffer_type = GL_UNSIGNED_INT_8_8_8_8_REV;
 #else
         *gl_buffer_type = GL_UNSIGNED_BYTE;
@@ -208,10 +217,12 @@ static void create_texture_if_needed(int width, int height) {
     //g_frame_texture.opengl_context_stamp = g_fs_ml_opengl_context_stamp;
     //fs_emu_set_texture(&g_frame_texture);
     fs_gl_bind_texture(g_frame_texture);
+#if 0
     // with high quality borders, there should be no reason to initialize
     // the texture to black
+    void *data = g_malloc0(g_frame_texture_width * g_frame_texture_height * 4);
+#endif
     void *data = NULL;
-    //void *data = fs_malloc0(g_frame_texture_width * g_frame_texture_height * 4);
     fs_gl_unpack_row_length(0);
     int gl_buffer_format = 0;
     int gl_buffer_type = 0;
@@ -220,9 +231,11 @@ static void create_texture_if_needed(int width, int height) {
             g_frame_texture_width, g_frame_texture_height, 0,
             gl_buffer_format, gl_buffer_type, data);
     CHECK_GL_ERROR();
+#if 0
     if (data) {
         free(data);
     }
+#endif
     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, g_texture_filter);
     CHECK_GL_ERROR();
     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, g_texture_filter);
@@ -336,7 +349,7 @@ static void save_screenshot(const char *path, int cx, int cy, int cw, int ch,
         uint8_t *op = out_data + y * cw * 3;
         if (frame_format == FS_EMU_VIDEO_FORMAT_BGRA) {
             for (int x = 0; x < row_len; x += frame_bpp) {
-#ifdef __BIG_ENDIAN__
+#ifdef WORDS_BIGENDIAN
                 *op++ = ip[x + 1];
                 *op++ = ip[x + 2];
                 *op++ = ip[x + 3];
@@ -465,18 +478,19 @@ static int update_texture() {
         static char* screenshots_prefix = NULL;
         static int screenshots_mask = 7;
         if (screenshots_dir == NULL) {
-            char *path = fs_config_get_string("screenshots_output_dir");
+            char *path = fs_config_get_string(OPTION_SCREENSHOTS_OUTPUT_DIR);
             if (path) {
+                path = fs_emu_path_expand_and_free(path);
                 if (fs_path_exists(path)) {
                     screenshots_dir = path;
                 }
                 else {
                     fs_emu_warning("Directory does not exist: %s", path);
-                    free(path);
+                    g_free(path);
                 }
             }
             if (!screenshots_dir) {
-                screenshots_dir = fs_strdup(fs_get_desktop_dir());
+                screenshots_dir = g_strdup(fs_get_desktop_dir());
             }
             screenshots_mask = fs_config_get_int("screenshots_output_mask");
             if (screenshots_mask == FS_CONFIG_NONE) {
@@ -486,7 +500,7 @@ static int update_texture() {
             screenshots_prefix = fs_config_get_string(
                     "screenshots_output_prefix");
             if (!screenshots_prefix) {
-                screenshots_prefix = fs_strdup("fs-uae");
+                screenshots_prefix = g_strdup("fs-uae");
             }
         }
 
@@ -506,37 +520,37 @@ static int update_texture() {
             strftime(strbuf, 20, "%y%m%d%H%M", tm_p);
             if (strcmp(strbuf, laststrbuf) != 0) {
                 count = 0;
-                strncpy(laststrbuf, strbuf, 20);
+                memcpy(laststrbuf, strbuf, 20);
             }
 
             count += 1;
             total_count += 1;
 
             if (screenshots_mask & 1) {
-                name = fs_strdup_printf("%s-%s-%s-%02d.png",
+                name = g_strdup_printf("%s-%s-%s-%02d.png",
                         screenshots_prefix, "full", strbuf, count);
-                path = fs_path_join(screenshots_dir, name, NULL);
+                path = g_build_filename(screenshots_dir, name, NULL);
                 save_screenshot(path, 0, 0, width, height, count, frame,
                         width, height, bpp);
-                free(path);
-                free(name);
+                g_free(path);
+                g_free(name);
             }
             if (screenshots_mask & 2) {
-                name = fs_strdup_printf("%s-%s-%s-%02d.png",
+                name = g_strdup_printf("%s-%s-%s-%02d.png",
                         screenshots_prefix, "crop", strbuf, count);
-                path = fs_path_join(screenshots_dir, name, NULL);
+                path = g_build_filename(screenshots_dir, name, NULL);
                 save_screenshot(path, g_crop.x, g_crop.y, g_crop.w, g_crop.h,
                         count, frame, width, height, bpp);
-                free(path);
-                free(name);
+                g_free(path);
+                g_free(name);
             }
             if (screenshots_mask & 4) {
-                name = fs_strdup_printf("%s-%s-%s-%02d.png",
+                name = g_strdup_printf("%s-%s-%s-%02d.png",
                         screenshots_prefix, "real", strbuf, count);
-                path = fs_path_join(screenshots_dir, name, NULL);
+                path = g_build_filename(screenshots_dir, name, NULL);
                 fs_ml_video_screenshot(path);
-                free(path);
-                free(name);
+                g_free(path);
+                g_free(name);
             }
             g_fs_emu_screenshot++;
         }
@@ -568,10 +582,16 @@ static int update_texture() {
         upload_h = buffer->height;
     }
 
-    if (g_fs_emu_scanlines &&
+    int filter = 0;
+    if (0) {
+        filter = 1;
+    }
+    else if (g_fs_emu_scanlines &&
             (buffer->flags & FS_EMU_NO_SCANLINES_FLAG) == 0) {
-        //printf("w %d h %d new frame? %d\n", buffer->width, buffer->height,
-        //        is_new_frame);
+        filter = 2;
+    }
+
+    if (filter) {
         if (is_new_frame) {
             if (g_scanline_buffer_width != buffer->width ||
                     g_scanline_buffer_height != buffer->height) {
@@ -582,9 +602,16 @@ static int update_texture() {
                 g_scanline_buffer_width = buffer->width;
                 g_scanline_buffer_height = buffer->height;
             }
-            fs_emu_render_scanlines(g_scanline_buffer, buffer,
-                    upload_x, upload_y, upload_w, upload_h,
-                    g_fs_emu_scanlines_dark, g_fs_emu_scanlines_light);
+            if (filter == 2) {
+                fs_emu_scanline_filter(g_scanline_buffer, buffer,
+                        upload_x, upload_y, upload_w, upload_h,
+                        g_fs_emu_scanlines_dark, g_fs_emu_scanlines_light);
+            }
+            else {
+                fs_emu_2xcolor_filter(g_scanline_buffer, buffer,
+                        upload_x, upload_y, upload_w, upload_h,
+                        g_fs_emu_scanlines_dark, g_fs_emu_scanlines_light);                
+            }
         }
         if (g_scanline_buffer) {
             frame = g_scanline_buffer;
@@ -630,7 +657,7 @@ static int update_texture() {
         //printf("%d %d\n", black_left, black_right);
         int black_width = black_right - black_left;
         if (black_width > 0) {
-            void *black_data = fs_malloc0(
+            void *black_data = g_malloc0(
                     black_width * g_frame_texture_height * bpp);
             //memset(black_data, 0xff, black_width * g_frame_texture_height * bpp);
             //printf("black1 w %d h %d\n", black_width, g_frame_texture_height);
@@ -651,7 +678,7 @@ static int update_texture() {
         int black_height = black_bottom - black_top;
         //printf("---- %d %d\n", black_top, black_bottom);
         if (black_height > 0) {
-            void *black_data = fs_malloc0(
+            void *black_data = g_malloc0(
                     upload_w * black_height * bpp);
             //memset(black_data, 0xff, upload_w * black_height * bpp);
             //printf("black2 w %d h %d\n", upload_w, black_height);
@@ -793,8 +820,117 @@ static void render_quad(float x1, float y1, float x2, float y2, float s1,
     CHECK_GL_ERROR();
 }
 
+static void render_overlays(void) {
+    fs_mutex_lock(g_overlay_mutex);
+    for (int i = 0; i < FS_EMU_MAX_OVERLAYS; i++) {
+        int state = g_overlay_status[i].state;
+#if 0
+        if (g_overlay_status[i].render_status &&
+                g_overlay_status[i].was_disabled) {
+            show = 0;
+        }
+        else if (!g_overlay_status[i].render_status &&
+                g_overlay_status[i].was_enabled) {
+            show = 1;
+        }
+        g_overlay_status[i].was_enabled = 0;
+        g_overlay_status[i].was_disabled = 0;
+#endif
+        g_overlay_status[i].render_state = state;
+        if (state) {
+            g_overlay_status[i].render_status++;
+        }
+        else {
+            g_overlay_status[i].render_status = 0;
+        }
+    }
+    fs_mutex_unlock(g_overlay_mutex);
+
+    for (int i = 0; i < FS_EMU_MAX_OVERLAYS; i++) {
+        //if (g_overlay_status[i].render_status < 1) {
+        //    continue;
+        //}
+        int state = g_overlay_status[i].render_state;
+        if (state < 0) state = 0;
+        if (state > FS_EMU_MAX_OVERLAY_STATES - 1) {
+            state = FS_EMU_MAX_OVERLAY_STATES - 1;
+        }
+
+        fs_emu_texture *overlay_texture = \
+                g_fs_emu_theme.overlays[i].textures[state];
+        if (!overlay_texture) {
+            continue;
+        }
+
+        float w = g_fs_emu_theme.overlays[i].w;
+        float h = g_fs_emu_theme.overlays[i].h;
+        float x1 = g_fs_emu_theme.overlays[i].x;
+        if (g_fs_emu_theme.overlays[i].anchor & FS_EMU_ANCHOR_RIGHT_BIT) {
+            x1 = x1 + 1.0 - w;
+        }
+        float y1 = g_fs_emu_theme.overlays[i].y;
+        if (g_fs_emu_theme.overlays[i].anchor & FS_EMU_ANCHOR_BOTTOM_BIT) {
+            y1 = y1 + 1.0 - h;
+        }
+
+        float x2 = x1 + w;
+        float y2 = y1 + h;
+
+        x1 = -1.0 + 2.0 * x1;
+        x2 = -1.0 + 2.0 * x2;
+        y1 =  1.0 - 2.0 * y1;
+        y2 =  1.0 - 2.0 * y2;
+
+        fs_gl_blending(1);
+        fs_gl_texturing(1);
+        fs_gl_color4f(1.0, 1.0, 1.0, 1.0);
+        fs_emu_set_texture(overlay_texture);
+
+#ifdef USE_GLES
+        GLfloat tex[] = {
+            0.0, 0.0,
+            1.0, 0.0,
+            1.0, 1.0,
+            0.0, 1.0
+        };
+        GLfloat vert[] = {
+            x1, y1,
+            x2, y1,
+            x2, y2,
+            x1, y2
+        };
+
+        glEnableClientState(GL_VERTEX_ARRAY);
+        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
+
+        glVertexPointer(2, GL_FLOAT, 0, vert);
+        glTexCoordPointer(2, GL_FLOAT, 0, tex);
+        glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
+
+        glDisableClientState(GL_VERTEX_ARRAY);
+        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
+#else
+        glBegin(GL_QUADS);
+        glTexCoord2f(0.0, 0.0); glVertex2f(x1, y1);
+        glTexCoord2f(1.0, 0.0); glVertex2f(x2, y1);
+        glTexCoord2f(1.0, 1.0); glVertex2f(x2, y2);
+        glTexCoord2f(0.0, 1.0); glVertex2f(x1, y2);
+        glEnd();
+#endif
+    }
+}
+
+#include <fs/emu/hacks.h>
+
+double fs_emu_video_scale_x = 0.0;
+double fs_emu_video_scale_y = 0.0;
+double fs_emu_video_offset_x = 1.0;
+double fs_emu_video_offset_y = 1.0;
+
 static void render_frame(double alpha, int perspective) {
+#ifdef WITH_LUA
     fs_emu_lua_run_handler("on_fs_emu_render_frame");
+#endif
 
 #ifdef WITH_LUA
 
@@ -825,12 +961,21 @@ static void render_frame(double alpha, int perspective) {
 
 #endif
 
+    int fox = g_fs_emu_theme.display_x;
+    int foy = g_fs_emu_theme.display_y;
+    int fow = g_fs_emu_theme.display_w;
+    int foh = g_fs_emu_theme.display_h;
+    int frame_override = fow != g_fs_emu_theme.width ||
+            foh != g_fs_emu_theme.height;
     if (g_frame_override) {
-        int x = g_frame_override_x;
-        int y = g_frame_override_y;
-        int w = g_frame_override_w;
-        int h = g_frame_override_h;
+        fox = g_frame_override_x;
+        foy = g_frame_override_y;
+        fow = g_frame_override_w;
+        foh = g_frame_override_h;
+        frame_override = 1;
+    }
 
+    if (frame_override) {
         float color = 1.0;
         if (g_frame_texture == 0) {
             // texture has not been created yet
@@ -844,10 +989,10 @@ static void render_frame(double alpha, int perspective) {
         fs_gl_blending(0);
         fs_gl_color4f(color, color, color, 1.0);
 
-        float x1 = -1.0 + x * 2.0 / g_l_scale_x;
-        float x2 = -1.0 + (x + w) * 2.0 / g_l_scale_x;
-        float y2 = 1.0 - y * 2.0 / g_l_scale_y;
-        float y1 = 1.0 - (y + h) * 2.0 / g_l_scale_y;
+        float x1 = -1.0 + fox * 2.0 / g_l_scale_x;
+        float x2 = -1.0 + (fox + fow) * 2.0 / g_l_scale_x;
+        float y2 = 1.0 - foy * 2.0 / g_l_scale_y;
+        float y1 = 1.0 - (foy + foh) * 2.0 / g_l_scale_y;
 
         double s1 = 0.0;
         double t1 = 0.0;
@@ -878,6 +1023,7 @@ static void render_frame(double alpha, int perspective) {
             render_quad(x1, y1, x2, y2, s1, t1, s2, t2);
         }
 
+        render_overlays();
         return;
     }
 
@@ -929,6 +1075,12 @@ static void render_frame(double alpha, int perspective) {
 
     //printf("w %d h %d x %d y %d\n", output_w, output_h, offset_x, offset_y);
 
+    fs_emu_video_scale_x = (double) input_w / output_w;
+    fs_emu_video_scale_y = (double) input_h / output_h;
+
+    fs_emu_video_offset_x = offset_x - g_crop.x / fs_emu_video_scale_x;
+    fs_emu_video_offset_y = offset_y - g_crop.y / fs_emu_video_scale_y;
+
     double s1 = 0.0;
     double t1 = 0.0;
     double s2 = (double) g_crop.w / g_frame_texture_width;
@@ -1344,104 +1496,7 @@ static void render_frame(double alpha, int perspective) {
         glEnd();
 #endif
     }
-
-    fs_mutex_lock(g_overlay_mutex);
-    for (int i = 0; i < FS_EMU_MAX_OVERLAYS; i++) {
-        int state = g_overlay_status[i].state;
-#if 0
-        if (g_overlay_status[i].render_status &&
-                g_overlay_status[i].was_disabled) {
-            show = 0;
-        }
-        else if (!g_overlay_status[i].render_status &&
-                g_overlay_status[i].was_enabled) {
-            show = 1;
-        }
-        g_overlay_status[i].was_enabled = 0;
-        g_overlay_status[i].was_disabled = 0;
-#endif
-        g_overlay_status[i].render_state = state;
-        if (state) {
-            g_overlay_status[i].render_status++;
-        }
-        else {
-            g_overlay_status[i].render_status = 0;
-        }
-    }
-    fs_mutex_unlock(g_overlay_mutex);
-
-    for (int i = 0; i < FS_EMU_MAX_OVERLAYS; i++) {
-        //if (g_overlay_status[i].render_status < 1) {
-        //    continue;
-        //}
-        int state = g_overlay_status[i].render_state;
-        if (state < 0) state = 0;
-        if (state > FS_EMU_MAX_OVERLAY_STATES - 1) {
-            state = FS_EMU_MAX_OVERLAY_STATES - 1;
-        }
-
-        fs_emu_texture *overlay_texture = \
-                g_fs_emu_theme.overlays[i].textures[state];
-        if (!overlay_texture) {
-            continue;
-        }
-
-        float w = g_fs_emu_theme.overlays[i].w;
-        float h = g_fs_emu_theme.overlays[i].h;
-        float x1 = g_fs_emu_theme.overlays[i].x;
-        if (g_fs_emu_theme.overlays[i].anchor & FS_EMU_ANCHOR_RIGHT_BIT) {
-            x1 = x1 + 1.0 - w;
-        }
-        float y1 = g_fs_emu_theme.overlays[i].y;
-        if (g_fs_emu_theme.overlays[i].anchor & FS_EMU_ANCHOR_BOTTOM_BIT) {
-            y1 = y1 + 1.0 - h;
-        }
-
-        float x2 = x1 + w;
-        float y2 = y1 + h;
-
-        x1 = -1.0 + 2.0 * x1;
-        x2 = -1.0 + 2.0 * x2;
-        y1 =  1.0 - 2.0 * y1;
-        y2 =  1.0 - 2.0 * y2;
-
-        fs_gl_blending(1);
-        fs_gl_texturing(1);
-        fs_gl_color4f(1.0, 1.0, 1.0, 1.0);
-        fs_emu_set_texture(overlay_texture);
-
-#ifdef USE_GLES
-        GLfloat tex[] = {
-            0.0, 0.0,
-            1.0, 0.0,
-            1.0, 1.0,
-            0.0, 1.0
-        };
-        GLfloat vert[] = {
-            x1, y1,
-            x2, y1,
-            x2, y2,
-            x1, y2
-        };
-    
-        glEnableClientState(GL_VERTEX_ARRAY);
-        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
-    
-        glVertexPointer(2, GL_FLOAT, 0, vert);
-        glTexCoordPointer(2, GL_FLOAT, 0, tex);
-        glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
-    
-        glDisableClientState(GL_VERTEX_ARRAY);
-        glDisableClientState(GL_TEXTURE_COORD_ARRAY);        
-#else        
-        glBegin(GL_QUADS);
-        glTexCoord2f(0.0, 0.0); glVertex2f(x1, y1);
-        glTexCoord2f(1.0, 0.0); glVertex2f(x2, y1);
-        glTexCoord2f(1.0, 1.0); glVertex2f(x2, y2);
-        glTexCoord2f(0.0, 1.0); glVertex2f(x1, y2);
-        glEnd();
-#endif
-    }
+    render_overlays();
 }
 
 static void render_glow(double opacity) {
@@ -1660,12 +1715,49 @@ static void render_glow(double opacity) {
     //printf("--- render glow done ---\n");
 }
 
+static void render_fade_overlay(double alpha) {
+    // draw fading effect
+    fs_gl_viewport(0, 0, fs_ml_video_width(), fs_ml_video_height());
+    fs_gl_ortho_hd();
+    fs_gl_blending(1);
+    fs_gl_texturing(0);
+    fs_gl_color4f(g_fs_emu_theme.fade_color[0] * alpha,
+            g_fs_emu_theme.fade_color[1] * alpha,
+            g_fs_emu_theme.fade_color[2] * alpha, alpha);
+
+#ifdef USE_GLES
+    GLfloat vert[] = {
+        0, 0,
+        1920, 0,
+        1920, 1080,
+        0, 1080
+    };
+
+    glEnableClientState(GL_VERTEX_ARRAY);
+
+    glVertexPointer(2, GL_FLOAT, 0, vert);
+    glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
+
+    glDisableClientState(GL_VERTEX_ARRAY);
+#else
+    glBegin(GL_QUADS);
+    glVertex2f(0, 0);
+    glVertex2f(1920, 0);
+    glVertex2f(1920, 1080);
+    glVertex2f(0, 1080);
+    glEnd();
+#endif
+
+    CHECK_GL_ERROR();
+}
+
 /**
  * This function is called at the end of the frame rendering function
  */
 static void handle_quit_sequence() {
     int fade_time = fs_config_get_int_clamped("fade_out_duration", 0, 10000);
     if (fade_time == FS_CONFIG_NONE) {
+        // fade out over 750ms
         fade_time = 750;
     }
     fade_time = fade_time * 1000;
@@ -1685,45 +1777,14 @@ static void handle_quit_sequence() {
         exit(1);
     }
 
-    // fade out over 750ms
-    float fade = (1.0 * dt) / fade_time;
-    if (fade > 1.0) {
-        fade = 1.0;
+
+    if (fade_time > 0) {
+        float fade = (1.0 * dt) / fade_time;
+        if (fade > 1.0) {
+            fade = 1.0;
+        }
+        render_fade_overlay(fade);
     }
-
-    // draw fading effect
-    fs_gl_viewport(0, 0, fs_ml_video_width(), fs_ml_video_height());
-    fs_gl_ortho_hd();
-    fs_gl_blending(1);
-    fs_gl_texturing(0);
-    fs_gl_color4f(g_fs_emu_theme.fade_color[0] * fade,
-            g_fs_emu_theme.fade_color[1] * fade,
-            g_fs_emu_theme.fade_color[2] * fade, fade);
-
-#ifdef USE_GLES
-    GLfloat vert[] = {
-        0, 0,
-        1920, 0,
-        1920, 1080,
-        0, 1080
-    };
-    
-    glEnableClientState(GL_VERTEX_ARRAY);
-    
-    glVertexPointer(2, GL_FLOAT, 0, vert);
-    glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
-    
-    glDisableClientState(GL_VERTEX_ARRAY);
-#else        
-    glBegin(GL_QUADS);
-    glVertex2f(0, 0);
-    glVertex2f(1920, 0);
-    glVertex2f(1920, 1080);
-    glVertex2f(0, 1080);
-    glEnd();
-#endif
-    
-    CHECK_GL_ERROR();
 }
 
 int fs_emu_video_update_function() {
@@ -2016,22 +2077,22 @@ void fs_emu_video_render_function() {
             int y = 1038;
 
             int rendered_tag = 0;
-            if (player->tag && player->tag[0]) {
-                str = fs_strdup_printf("%s", player->tag);
+            if (player->tag[0]) {
+                str = g_strdup_printf("%s", player->tag);
                 fs_emu_font_render(menu_font, str, x, y,
                         1.0, 1.0, 1.0, 1.0);
                 free(str);
                 rendered_tag = 1;
             }
             if (rendered_tag || player->ping) {
-                str = fs_strdup_printf("%03d", player->ping);
+                str = g_strdup_printf("%03d", player->ping);
                 fs_emu_font_render(menu_font, str, x + 100, y,
                         1.0, 1.0, 1.0, 1.0);
                 free(str);
             }
             /*
             if (rendered_tag || player->lag) {
-                str = fs_strdup_printf("%03d", player->lag);
+                str = g_strdup_printf("%03d", player->lag);
                 fs_emu_font_render(menu_font, str, x + 200, y,
                         1.0, 1.0, 1.0, 1.0);
                 free(str);
@@ -2060,7 +2121,7 @@ void fs_emu_video_render_function() {
         static GLuint debug_texture = 0;
         static uint32_t *debug_texture_data = NULL;
         if (debug_texture == 0) {
-            debug_texture_data = fs_malloc0(256 * 256 * 4);
+            debug_texture_data = g_malloc0(256 * 256 * 4);
             glGenTextures(1, &debug_texture);
             CHECK_GL_ERROR();
             fs_gl_bind_texture(debug_texture);
@@ -2141,38 +2202,49 @@ void fs_emu_video_render_function() {
         char *str;
 
         /*
-        str = fs_strdup_printf("%d", fs_emu_get_audio_frequency());
+        str = g_strdup_printf("%d", fs_emu_get_audio_frequency());
         fs_emu_font_render(menu_font, str, 1920 / 2 + 20, 3,
                 1.0, 1.0, 1.0, 1.0);
         free(str);
         */
 
-        str = fs_strdup_printf("%0.1f",
-                fs_emu_audio_get_measured_avg_buffer_fill(0) / 1000.0);
+        static double avg_emu = 0.0;
+        static double avg_sys = 0.0;
+        static double avg_fill = 0.0;
+        static int avg_throttle = 0;
+        if (avg_throttle == 0) {
+            avg_emu = fs_emu_get_average_emu_fps();
+            avg_sys = fs_emu_get_average_sys_fps();
+            avg_fill = fs_emu_audio_get_measured_avg_buffer_fill(0);
+            avg_throttle = 10;
+        }
+        avg_throttle -= 1;
+
+        str = g_strdup_printf("%0.1f", avg_fill / 1000.0);
         fs_emu_font_render(menu_font, str, 1920 / 2 + 220, 3,
                 1.0, 1.0, 1.0, 1.0);
         free(str);
-        str = fs_strdup_printf("%d", g_fs_emu_audio_buffer_underruns);
+        str = g_strdup_printf("%d", g_fs_emu_audio_buffer_underruns);
         fs_emu_font_render(menu_font, str, 1920 / 2 + 420, 3,
                 1.0, 1.0, 1.0, 1.0);
         free(str);
 
         fs_emu_font_render(menu_font, "EMU", 20, 3, 1.0, 1.0, 1.0, 1.0);
-        str = fs_strdup_printf("%0.1f", fs_emu_get_average_emu_fps());
+        str = g_strdup_printf("%0.2f", avg_emu);
         fs_emu_font_render(menu_font, str, 220, 3, 1.0, 1.0, 1.0, 1.0);
         free(str);
-        str = fs_strdup_printf("%d", g_fs_emu_lost_frames);
+        str = g_strdup_printf("%d", g_fs_emu_lost_frames);
         fs_emu_font_render(menu_font, str, 420, 3, 1.0, 1.0, 1.0, 1.0);
         free(str);
-        str = fs_strdup_printf("%d", g_fs_emu_repeated_frames);
+        str = g_strdup_printf("%d", g_fs_emu_repeated_frames);
         fs_emu_font_render(menu_font, str, 620, 3, 1.0, 1.0, 1.0, 1.0);
         free(str);
 
         fs_emu_font_render(menu_font, "SYS", 20, 140, 1.0, 1.0, 1.0, 1.0);
-        str = fs_strdup_printf("%0.1f", fs_emu_get_average_sys_fps());
+        str = g_strdup_printf("%0.2f", avg_sys);
         fs_emu_font_render(menu_font, str, 220, 140, 1.0, 1.0, 1.0, 1.0);
         free(str);
-        str = fs_strdup_printf("%d", g_fs_emu_lost_vblanks);
+        str = g_strdup_printf("%d", g_fs_emu_lost_vblanks);
         fs_emu_font_render(menu_font, str, 420, 140, 1.0, 1.0, 1.0, 1.0);
         free(str);
 
@@ -2184,6 +2256,28 @@ void fs_emu_video_render_function() {
         handle_quit_sequence();
     }
 
+    static int64_t start_time = -1;
+    if (start_time == -1) {
+        start_time = fs_emu_monotonic_time();
+    }
+    static int fade_time = FS_CONFIG_NONE;
+    if (fade_time == FS_CONFIG_NONE) {
+        fade_time = fs_config_get_int_clamped("fade_in_duration", 0, 10000);
+        if (fade_time == FS_CONFIG_NONE) {
+            fade_time = 0;
+        }
+        fade_time = fade_time * 1000;
+    }
+
+    uint64_t dt = fs_emu_monotonic_time() - start_time;
+    if (dt < fade_time) {
+        float fade = 1.0 - (1.0 * dt) / fade_time;
+        if (fade < 0.0) {
+            fade = 0.0;
+        }
+        render_fade_overlay(fade);
+    }
+
     fs_emu_video_render_mutex_unlock();
 }
 
@@ -2192,7 +2286,7 @@ void fs_emu_video_render_debug_info(uint32_t *texture) {
     int x;
     //, y;
     int y1;
-    fs_list *link;
+    GList *link;
     uint32_t color = 0x80404080;
     fs_gl_ortho_hd();
     fs_gl_blending(TRUE);
@@ -2217,7 +2311,7 @@ void fs_emu_video_render_debug_info(uint32_t *texture) {
     x = 127;
     y1 = 128;
     color = 0x80404080;
-    link = fs_queue_peek_head_link(g_fs_emu_sys_frame_times.queue);
+    link = g_queue_peek_head_link(g_fs_emu_sys_frame_times.queue);
     while (link) {
         int val = FS_POINTER_TO_INT(link->data);
         //int x2 = x - 8;
@@ -2237,7 +2331,7 @@ void fs_emu_video_render_debug_info(uint32_t *texture) {
     x = 127;
     y1 = 0;
     color = 0x80205080;
-    link = fs_queue_peek_head_link(g_fs_emu_emu_frame_times.queue);
+    link = g_queue_peek_head_link(g_fs_emu_emu_frame_times.queue);
     while (link) {
         int val = FS_POINTER_TO_INT(link->data);
         //int x2 = x - 8;
@@ -2257,7 +2351,7 @@ void fs_emu_video_render_debug_info(uint32_t *texture) {
     x = 127;
     y1 = 0;
     color = 0x80008080;
-    link = fs_queue_peek_head_link(g_fs_emu_emu2_frame_times.queue);
+    link = g_queue_peek_head_link(g_fs_emu_emu2_frame_times.queue);
     while (link) {
         int val = FS_POINTER_TO_INT(link->data);
         //int x2 = x - 8;
@@ -2380,7 +2474,7 @@ void fs_emu_init_render() {
     }
 
     const char *ccstr = fs_config_get_const_string("texture_filter");
-    if (ccstr && fs_ascii_strcasecmp(ccstr, "nearest") == 0) {
+    if (ccstr && g_ascii_strcasecmp(ccstr, "nearest") == 0) {
         g_texture_filter = GL_NEAREST;
     }
 
