@@ -1,15 +1,21 @@
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
+#include <fs/emu.h>
 #include "xml_shader.h"
 
 #ifdef WITH_XML_SHADER
 
 #include <fs/ml/opengl.h>
-#include <fs/glee.h>
+// #include <GLee.h>
+#include <GL/glew.h>
 
 #include <fs/emu.h>
+#include <fs/glib.h>
 #include <fs/i18n.h>
-#include <fs/list.h>
 #include <fs/log.h>
-#include <fs/string.h>
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -49,13 +55,13 @@ typedef struct shader_pass {
 typedef struct fs_emu_shader {
     char *path;
     int ok;
-    fs_list *passes;
-    fs_list *current_shaders;
+    GList *passes;
+    GList *current_shaders;
 } fs_emu_shader;
 
 //static int g_shader_ok;
-//static fs_list *g_shader_passes = NULL;
-//static fs_list *g_current_shaders = NULL;
+//static GList *g_shader_passes = NULL;
+//static GList *g_current_shaders = NULL;
 static fs_emu_shader *g_active_shader = NULL;
 
 #define MAX_TEXT_SIZE 65536
@@ -114,7 +120,7 @@ static void on_start_element(GMarkupParseContext *context,
     const gchar **v = attribute_values;
     for (const gchar **k = attribute_names; *k; k++, v++) {
         if (strcmp(*k, "filter") == 0) {
-            data->filter = fs_strdup(*v);
+            data->filter = g_strdup(*v);
             data->flags |= A_FILTER;
         }
         else if (strcmp(*k, "size_x") == 0) {
@@ -178,7 +184,7 @@ static void on_text(GMarkupParseContext *context, const gchar *text,
     g_free(value);
 }
 
-static void log_shader_error(int shader) {
+static void log_shader_log(int shader) {
     GLint info_length = 0;
     glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &info_length);
     if (info_length > 0) {
@@ -245,17 +251,16 @@ static void handle_element(parse_data *data, const char *element,
         GLint compile_status;
         glGetShaderiv(shader, GL_COMPILE_STATUS, &compile_status);
         CHECK_GL_ERROR();
+        log_shader_log(shader);
         if (compile_status == GL_FALSE) {
             fs_emu_warning("failed to compile vertex shader");
-            log_shader_error(shader);
             data->error = 1;
             return;
         }
         else {
             fs_log("compiled vertex shader successfully\n");
         }
-
-        data->shader->current_shaders = fs_list_append(
+        data->shader->current_shaders = g_list_append(
                 data->shader->current_shaders,
                 FS_UINT_TO_POINTER(shader));
         return;
@@ -368,9 +373,9 @@ static void handle_element(parse_data *data, const char *element,
     GLint compile_status;
     glGetShaderiv(shader, GL_COMPILE_STATUS, &compile_status);
     CHECK_GL_ERROR();
+    log_shader_log(shader);
     if (compile_status == GL_FALSE) {
         fs_emu_warning("failed to compile fragment shader");
-        log_shader_error(shader);
         data->error = 1;
         return;
     }
@@ -380,7 +385,7 @@ static void handle_element(parse_data *data, const char *element,
 
     // Add the shader handle to the current list of shaders.
 
-    data->shader->current_shaders = fs_list_append(
+    data->shader->current_shaders = g_list_append(
             data->shader->current_shaders,
             FS_UINT_TO_POINTER(shader));
 
@@ -395,16 +400,16 @@ static void handle_element(parse_data *data, const char *element,
         return;
     }
 
-    fs_list *link = data->shader->current_shaders;
+    GList *link = data->shader->current_shaders;
     while (link) {
         GLuint shader = GPOINTER_TO_UINT(link->data);
         glAttachShader(program, shader);
         // schedule shader for deletion so it will be deleted when the
         // program is deleted
         glDeleteShader(shader);
-        fs_list *temp = link;
+        GList *temp = link;
         link = link->next;
-        fs_list_free_1(temp);
+        g_list_free_1(temp);
     }
     data->shader->current_shaders = NULL;
 
@@ -454,7 +459,7 @@ static void handle_element(parse_data *data, const char *element,
         }
     }
     // Add the shader pass to the list of shader passes.
-    data->shader->passes = fs_list_append(data->shader->passes, pass);
+    data->shader->passes = g_list_append(data->shader->passes, pass);
 }
 
 static void on_end_element(GMarkupParseContext *context,
@@ -501,7 +506,7 @@ static void context_notification_handler(int notification, void *data) {
     fs_emu_shader *shader = (fs_emu_shader *) data;
     if (notification == FS_GL_CONTEXT_DESTROY) {
         fs_log("FS_GL_CONTEXT_DESTROY handler for shader\n");
-        fs_list *link = shader->passes;
+        GList *link = shader->passes;
         if (shader->passes == NULL) {
             return;
         }
@@ -511,9 +516,9 @@ static void context_notification_handler(int notification, void *data) {
             shader_pass *pass = link->data;
             glDeleteProgram(pass->program);
             g_free(pass);
-            fs_list *delete_link = link;
+            GList *delete_link = link;
             link = link->next;
-            fs_list_free_1(delete_link);
+            g_list_free_1(delete_link);
         }
         recreate = 1;
         shader->passes = NULL;
@@ -534,7 +539,7 @@ void fs_emu_load_shader(fs_emu_shader *shader) {
     GMarkupParseContext *context = g_markup_parse_context_new(
             &counter_subparser, G_MARKUP_TREAT_CDATA_AS_TEXT, data, NULL);
 
-    FILE *f = fs_fopen(data->shader->path, "rb");
+    FILE *f = g_fopen(data->shader->path, "rb");
     if (f == NULL) {
         fs_log("could not open shader file\n");
         return;
@@ -577,17 +582,17 @@ void fs_emu_load_shader(fs_emu_shader *shader) {
 }
 
 static char *find_shader(const char *name) {
-    //char *path = fs_strdup(name);
+    //char *path = g_strdup(name);
     fs_log("checking shader %s\n", name);
     if (fs_path_exists(name)) {
-        return fs_strdup(name);
+        return g_strdup(name);
     }
-    char *name2 = fs_strconcat(name, ".shader", NULL);
-    char *path = fs_path_join("shaders", name2, NULL);
-    free(name2);
+    char *name2 = g_strconcat(name, ".shader", NULL);
+    char *path = g_build_filename("shaders", name2, NULL);
+    g_free(name2);
     fs_log("checking shader (share)/%s\n", path);
     char *path2 = fs_get_program_data_file(path);
-    free(path);
+    g_free(path);
     if (path2) {
         return path2;
     }
@@ -1123,7 +1128,7 @@ int fs_emu_xml_shader_render(int texture, int texture_width,
     debug_printf("     texture: %d %d\n", g_cur_texture_w, g_cur_texture_h);
 
     // For each shader pass in the list of shader passes...
-    fs_list *link = g_active_shader->passes;
+    GList *link = g_active_shader->passes;
     int first = 1;
     while (link) {
         shader_pass *pass = link->data;
