@@ -1,13 +1,25 @@
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
+// FIXME: make libfsml independent of libfsmeu
+#include "../emu/util.h"
+#include "../emu/video.h"
+
 #include <stdlib.h>
 #include <string.h>
 #include <stdio.h>
 #include <stddef.h>
 
+#ifdef USE_SDL2
+#define USE_SDL
+#endif
+
 #ifdef USE_SDL
 #include <SDL.h>
 #endif
 
-#include <fs/config.h>
+#include <fs/conf.h>
 #include <fs/i18n.h>
 #include <fs/image.h>
 
@@ -20,9 +32,6 @@
 #include "ml_internal.h"
 
 #include <fs/thread.h>
-// FIXME: make libfsml independent of libfsmeu
-#include "../emu/util.h"
-#include "../emu/video.h"
 
 static int g_vblank_count = 0;
 static int64_t g_measured_vblank_time = 0;
@@ -132,17 +141,17 @@ static void save_screenshot_of_opengl_framebuffer(const char *path) {
 #endif
     char strbuf[20];
     strftime(strbuf, 20, "%Y-%m-%d-%H-%M", tm_p);
-    char *name = fs_strdup_printf("%s-%s-%03d.png",
+    char *name = g_strdup_printf("%s-%s-%03d.png",
             g_fs_ml_video_screenshots_prefix,
             strbuf, g_fs_ml_video_screenshot);
-    char *path = fs_path_join(g_fs_ml_video_screenshots_dir, name, NULL);
+    char *path = g_build_filename(g_fs_ml_video_screenshots_dir, name, NULL);
 #endif
     fs_log("writing screenshot to %s\n", path);
 
     int w = fs_ml_video_width();
     int h = fs_ml_video_height();
     fs_log("reading opengl frame buffer (%d x %d)\n", w, h);
-    void *out_data = malloc(w * h * 4);
+    void *out_data = g_malloc(w * h * 4);
 
     // when using GL_RGB, remeber to temporarily set GL_UNPACK_ALIGNMENT so
     // all rows will be contiguous (the OpenGL default is to align rows on
@@ -155,7 +164,7 @@ static void save_screenshot_of_opengl_framebuffer(const char *path) {
 
     // flip image vertically
     int stride = w * 4;
-    void *tmp = malloc(stride);
+    void *tmp = g_malloc(stride);
     void *line1 = out_data;
     void *line2 = out_data + stride * (h - 1);
     for (int i = 0; i < h / 2; i++) {
@@ -165,7 +174,7 @@ static void save_screenshot_of_opengl_framebuffer(const char *path) {
         line1 += stride;
         line2 -= stride;
     }
-    free(tmp);
+    g_free(tmp);
 
     int result = fs_image_save_data(path, out_data, w, h, 4);
     if (result) {
@@ -174,10 +183,10 @@ static void save_screenshot_of_opengl_framebuffer(const char *path) {
     else {
         fs_log("error saving screenshot\n");
     }
-    free(out_data);
+    g_free(out_data);
 #if 0
-    free(name);
-    free(path);
+    g_free(name);
+    g_free(path);
 #endif
 }
 
@@ -215,9 +224,15 @@ static void render_frame() {
     }
 }
 
+#ifdef USE_SDL2
+extern SDL_Window* g_fs_ml_window;
+#endif
+
 static void swap_opengl_buffers() {
     //int64_t t1 = fs_get_monotonic_time();
-#ifdef USE_SDL
+#if defined(USE_SDL2)
+    SDL_GL_SwapWindow(g_fs_ml_window);
+#elif defined(USE_SDL)
     SDL_GL_SwapBuffers();
 #else
     printf("ERROR: no swap\n");
@@ -275,6 +290,7 @@ static void full_sleep_until_vsync() {
 }
 #endif
 
+#ifdef USE_GLEE
 typedef int64_t GLint64;
 typedef uint64_t GLuint64;
 typedef struct __GLsync *GLsync;
@@ -285,6 +301,7 @@ static GLenum (APIENTRYP glClientWaitSync)(GLsync sync, GLbitfield flags,
         GLuint64 timeout) = NULL;
 static GLsync (APIENTRYP glFenceSync)(GLenum condition, 
         GLbitfield flags) = NULL;
+#endif
 
 #define GL_OBJECT_TYPE                 0x9112
 #define GL_SYNC_CONDITION              0x9113
@@ -325,7 +342,7 @@ void * __GLeeGetProcAddress(const char *extname);
 static int g_sync_method = 0;
 
 static int check_sync_method(const char *a, const char *b) {
-    if (a && fs_ascii_strcasecmp(a, b) == 0) {
+    if (a && g_ascii_strcasecmp(a, b) == 0) {
         return 1;
     }
     return 0;
@@ -398,9 +415,11 @@ static void check_opengl_sync_capabilities() {
         }
         if (strstr(ext, "GL_ARB_sync") != NULL) {
             fs_log("GL_ARB_sync extension found\n");
+#ifdef USE_GLEE
             glFenceSync = __GLeeGetProcAddress("glFenceSync");
             glWaitSync = __GLeeGetProcAddress("glWaitSync");
             glClientWaitSync = __GLeeGetProcAddress("glClientWaitSync");
+#endif
             if (glFenceSync && glClientWaitSync) {
                 g_has_arb_sync = 1;
             }
@@ -611,18 +630,24 @@ void fs_ml_render_iteration() {
             // frames so there's no point waiting for them. Instead, we just
             // sleep a bit to throttle the frame rate for the quit animation
             fs_ml_usleep(10000);
-        }
-        else {
+        } else {
             // wait max 33 ms to allow the user interface to work even if
             // the emu hangs
-            int64_t dest_time = fs_get_real_time() + 33 * 1000;
+            // int64_t dest_time = fs_get_real_time() + 33 * 1000;
+            int64_t end_time = fs_condition_get_wait_end_time(33 * 1000);
+            int64_t check_time = 0;
 
             fs_mutex_lock(g_frame_available_mutex);
+            // fs_log("cond wait until %lld\n", end_time);
             while (g_rendered_frame == g_available_frame) {
-                fs_condition_timed_wait(g_frame_available_cond,
-                        g_frame_available_mutex, dest_time);
-                if (fs_get_real_time() >= dest_time) {
+                fs_condition_wait_until(
+                    g_frame_available_cond, g_frame_available_mutex, end_time);
+                check_time = fs_condition_get_wait_end_time(0);
+                if (check_time >= end_time) {
+                    // fs_log("timed out at %lld\n", check_time);
                     break;
+                } else {
+                    // fs_log("wake-up at %lld (end_time = %lld)\n", check_time, end_time);
                 }
             }
             fs_mutex_unlock(g_frame_available_mutex);
@@ -639,7 +664,7 @@ void fs_ml_render_iteration() {
         if (g_fs_ml_video_screenshot_path) {
             save_screenshot_of_opengl_framebuffer(
                     g_fs_ml_video_screenshot_path);
-            free(g_fs_ml_video_screenshot_path);
+            g_free(g_fs_ml_video_screenshot_path);
             g_fs_ml_video_screenshot_path = NULL;
         }
         fs_mutex_unlock(g_fs_ml_video_screenshot_mutex);
