@@ -9,8 +9,6 @@
 *
 */
 
-//#define ROMHACK
-//#define ROMHACK2
 //#define CDTV_SUB_DEBUG
 //#define CDTV_DEBUG
 //#define CDTV_DEBUG_CMD
@@ -32,6 +30,7 @@
 #include "a2091.h"
 #include "uae.h"
 #include "savestate.h"
+#include "scsi.h"
 
 /* DMAC CNTR bits. */
 #define CNTR_TCEN               (1<<7)
@@ -93,13 +92,6 @@ static int cdrom_command_cnt_in;
 static uae_u8 tp_a, tp_b, tp_c, tp_ad, tp_bd, tp_cd;
 static uae_u8 tp_imask, tp_cr, tp_air, tp_ilatch, tp_ilatch2;
 
-#ifdef ROMHACK
-#define ROM_VECTOR 0x2000
-#define ROM_OFFSET 0x2000
-static int rom_size, rom_mask;
-static uae_u8 *rom;
-#endif
-
 static void do_stch (void);
 
 static void INT2 (void)
@@ -517,7 +509,8 @@ static void cdrom_command_thread (uae_u8 b)
 		if (cdrom_command_cnt_in == 7) {
 			cdrom_command_accepted (0, s, &cdrom_command_cnt_in);
 			cd_finished = 1;
-			sleep_millis (500);
+			if (currprefs.cd_speed)
+				sleep_millis (500);
 			activate_stch = 1;
 		}
 		break;
@@ -648,6 +641,8 @@ static void dma_do_thread (void)
 		cdrom_offset / cdtv_sectorsize, dmac_acr, cnt, cdrom_length / 2);
 #endif
 	dma_wait += cnt * (uae_u64)312 * 50 / 75 + 1;
+	if (currprefs.cd_speed == 0)
+		dma_wait = 1;
 	while (cnt > 0 && dmac_dma) {
 		uae_u8 buffer[2352];
 		if (!didread || readsector != (cdrom_offset / cdtv_sectorsize)) {
@@ -1049,6 +1044,16 @@ static void dmac_start_dma (void)
 {
 	if (!(dmac_cntr & CNTR_PDMD)) { // non-scsi dma
 		write_comm_pipe_u32 (&requests, 0x0100, 1);
+	} else {
+		scsi_dmac_a2091_start_dma (wd_cdtv);
+	}
+}
+static void dmac_stop_dma (void)
+{
+	if (!(dmac_cntr & CNTR_PDMD)) { // non-scsi dma
+		;
+	} else {
+		scsi_dmac_a2091_stop_dma (wd_cdtv);
 	}
 }
 
@@ -1061,7 +1066,7 @@ static void checkint (void)
 {
 	int irq = 0;
 
-	if (currprefs.cs_cdtvscsi && (wdscsi_getauxstatus () & 0x80)) {
+	if (currprefs.cs_cdtvscsi && (wdscsi_getauxstatus (&wd_cdtv->wc) & 0x80)) {
 		dmac_istr |= ISTR_INTS;
 		if ((dmac_cntr & CNTR_INTEN) && (dmac_istr & ISTR_INTS))
 			irq = 1;
@@ -1092,7 +1097,7 @@ void CDTV_hsync_handler (void)
 {
 	static int subqcnt;
 
-	if (!currprefs.cs_cdtvcd || configured <= 0)
+	if (!currprefs.cs_cdtvcd || configured <= 0 || currprefs.cs_cdtvcr)
 		return;
 
 	cdtv_hsync++;
@@ -1130,6 +1135,7 @@ void CDTV_hsync_handler (void)
 		cd_finished = 1;
 		cd_paused = 0;
 		//cd_error = 1;
+		write_log (_T("audio finished\n"));
 		activate_stch = 1;
 	}
 
@@ -1203,28 +1209,17 @@ void CDTV_hsync_handler (void)
 
 static void do_stch (void)
 {
-	static int stch_cnt;
-
 	if ((tp_cr & 1) && !(tp_air & (1 << 2))) {
 		stch = 1;
 		activate_stch = 0;
 		tp_check_interrupts ();
 #ifdef CDTV_DEBUG
+		static int stch_cnt;
 		write_log (_T("STCH %d\n"), stch_cnt++);
 #endif
 	}
 }
 
-void bleh (void)
-{
-#if 0
-	cd_playing = cd_finished = cd_motor = cd_media = 1;
-	cd_isready = 0;
-	cd_playing = 0;
-	do_stch();
-#endif
-}
-
 static void cdtv_reset_int (void)
 {
 	write_log (_T("CDTV: reset\n"));
@@ -1236,6 +1231,7 @@ static void cdtv_reset_int (void)
 	cd_finished = 0;
 	cd_led = 0;
 	stch = 1;
+	frontpanel = 1;
 }
 
 static uae_u32 dmac_bget2 (uaecptr addr)
@@ -1249,16 +1245,6 @@ static uae_u32 dmac_bget2 (uaecptr addr)
 	if (addr >= 0xb0 && addr < 0xc0)
 		return tp_bget ((addr - 0xb0) / 2);
 
-#ifdef ROMHACK
-	if (addr >= ROM_OFFSET) {
-		if (rom) {
-			int off = addr & rom_mask;
-			return rom[off];
-		}
-		return 0;
-	}
-#endif
-
 	switch (addr)
 	{
 	case 0x41:
@@ -1272,11 +1258,11 @@ static uae_u32 dmac_bget2 (uaecptr addr)
 		break;
 	case 0x91:
 		if (currprefs.cs_cdtvscsi)
-			v = wdscsi_getauxstatus ();
+			v = wdscsi_getauxstatus (&wd_cdtv->wc);
 		break;
 	case 0x93:
 		if (currprefs.cs_cdtvscsi) {
-			v = wdscsi_get ();
+			v = wdscsi_get (&wd_cdtv->wc, wd_cdtv);
 			checkint ();
 		}
 		break;
@@ -1379,13 +1365,13 @@ static void dmac_bput2 (uaecptr addr, uae_u32 b)
 		break;
 	case 0x91:
 		if (currprefs.cs_cdtvscsi) {
-			wdscsi_sasr (b);
+			wdscsi_sasr (&wd_cdtv->wc, b);
 			checkint ();
 		}
 		break;
 	case 0x93:
 		if (currprefs.cs_cdtvscsi) {
-			wdscsi_put (b);
+			wdscsi_put (&wd_cdtv->wc, wd_cdtv, b);
 			checkint ();
 		}
 		break;
@@ -1401,7 +1387,10 @@ static void dmac_bput2 (uaecptr addr, uae_u32 b)
 		break;
 	case 0xe2:
 	case 0xe3:
-		dmac_dma = 0;
+		if (dmac_dma) {
+			dmac_dma = 0;
+			dmac_stop_dma ();
+		}
 		dma_finished = 0;
 		break;
 	case 0xe4:
@@ -1496,16 +1485,14 @@ static void REGPARAM2 dmac_bput (uaecptr addr, uae_u32 b)
 	addr &= 65535;
 	b &= 0xff;
 	if (addr == 0x48) {
-		map_banks (&dmac_bank, b, 0x10000 >> 16, 0x10000);
-		write_log (_T("CDTV DMAC autoconfigured at %02X0000\n"), b);
+		map_banks_z2 (&dmac_bank, b, 0x10000 >> 16);
 		configured = b;
-		expamem_next ();
+		expamem_next(&dmac_bank, NULL);
 		return;
 	}
 	if (addr == 0x4c) {
-		write_log (_T("CDTV DMAC AUTOCONFIG SHUT-UP!\n"));
 		configured = -1;
-		expamem_next ();
+		expamem_shutup(&dmac_bank);
 		return;
 	}
 	if (configured <= 0)
@@ -1544,11 +1531,6 @@ static uae_u32 REGPARAM2 dmac_wgeti (uaecptr addr)
 	uae_u32 v = 0xffff;
 #ifdef JIT
 	special_mem |= S_READ;
-#endif
-#ifdef ROMHACK
-	addr &= 65535;
-	if (addr >= ROM_OFFSET)
-		v = (rom[addr & rom_mask] << 8) | rom[(addr + 1) & rom_mask];
 #endif
 	return v;
 }
@@ -1557,10 +1539,6 @@ static uae_u32 REGPARAM2 dmac_lgeti (uaecptr addr)
 	uae_u32 v = 0xffff;
 #ifdef JIT
 	special_mem |= S_READ;
-#endif
-#ifdef ROMHACK
-	addr &= 65535;
-	v = (dmac_wgeti(addr) << 16) | dmac_wgeti(addr + 2);
 #endif
 	return v;
 }
@@ -1569,7 +1547,7 @@ static uae_u32 REGPARAM2 dmac_lgeti (uaecptr addr)
 addrbank dmac_bank = {
 	dmac_lget, dmac_wget, dmac_bget,
 	dmac_lput, dmac_wput, dmac_bput,
-	default_xlate, default_check, NULL, _T("CDTV DMAC/CD Controller"),
+	default_xlate, default_check, NULL, NULL, _T("CDTV DMAC/CD Controller"),
 	dmac_lgeti, dmac_wgeti, ABFLAG_IO
 };
 
@@ -1656,13 +1634,6 @@ uae_u8 cdtv_battram_read (int addr)
 	return v;
 }
 
-int cdtv_add_scsi_unit(int ch, TCHAR *path, int blocksize, int readonly,
-	TCHAR *devname, int sectors, int surfaces, int reserved,
-	int bootpri, TCHAR *filesys)
-{
-	return addscsi (ch, path, blocksize, readonly, devname, sectors, surfaces, reserved, bootpri, filesys, 1);
-}
-
 void cdtv_free (void)
 {
 	if (thread_alive > 0) {
@@ -1679,49 +1650,7 @@ void cdtv_free (void)
 	configured = 0;
 }
 
-
-#ifdef ROMHACK2
-extern uae_u8 *extendedkickmemory, *cardmemory;
-static void romhack (void)
-{
-	struct zfile *z;
-	int roms[5];
-	struct romlist *rl;
-	int rom_size;
-	uae_u8 *rom, *p;
-
-	extendedkickmemory[0x558c] = 0xff;
-
-	roms[0] = 55;
-	roms[1] = 54;
-	roms[2] = 53;
-	roms[3] = -1;
-
-	rl = getromlistbyids(roms);
-	if (rl) {
-		write_log (_T("A590/A2091 BOOT ROM '%s' %d.%d\n"), rl->path, rl->rd->ver, rl->rd->rev);
-		z = zfile_fopen(rl->path, "rb", ZFD_NORMAL);
-		if (z) {
-			rom_size = 16384;
-			rom = (uae_u8*)xmalloc (rom_size);
-			zfile_fread (rom, rom_size, 1, z);
-			rom[0x2071] = 0xe0; rom[0x2072] |= 0x40;
-			rom[0x2075] = 0xe0; rom[0x2076] |= 0x40;
-			rom[0x207d] = 0xe0; rom[0x207e] |= 0x40;
-			rom[0x2081] = 0xe0; rom[0x2082] |= 0x40;
-			rom[0x2085] = 0xe0; rom[0x2086] |= 0x40;
-			rom[0x207b] = 0x32;
-			p = cardmemory + 0x4000;
-			memcpy (p, rom + 0x2000, 0x2000);
-			memcpy (p + 0x2000, rom, 0x2000);
-		}
-		zfile_fclose(z);
-	}
-	//kickmemory[0x3592c] = 0xff;
-}
-#endif
-
-void cdtv_init (void)
+addrbank *cdtv_init (struct romconfig *rc)
 {
 	close_unit ();
 	if (!thread_alive) {
@@ -1748,8 +1677,8 @@ void cdtv_init (void)
 	ew (0x24, 0x00); /* ser.no. Byte 3 */
 
 	/* KS autoconfig handles the rest */
-	map_banks (&dmac_bank, 0xe80000 >> 16, 0x10000 >> 16, 0x10000);
 	if (!savestate_state) {
+		cdtv_reset_int ();
 		configured = 0;
 		tp_a = tp_b = tp_c = tp_ad = tp_bd = tp_cd = 0;
 		tp_imask = tp_cr = tp_air = tp_ilatch = 0;
@@ -1762,21 +1691,26 @@ void cdtv_init (void)
 	cdtv_battram_reset ();
 	open_unit ();
 	gui_flicker_led (LED_CD, 0, -1);
+	if (currprefs.cs_cdtvscsi) {
+		init_wd_scsi (wd_cdtv);
+		wd_cdtv->dmac_type = COMMODORE_DMAC;
+	}
+	return &dmac_bank;
 }
 
 void cdtv_check_banks (void)
 {
 	if (configured > 0)
-		map_banks (&dmac_bank, configured, 0x10000 >> 16, 0x10000);
+		map_banks_z2 (&dmac_bank, configured, 0x10000 >> 16);
 }
 
 #ifdef SAVESTATE
 
-uae_u8 *save_dmac (int *len, uae_u8 *dstptr)
+uae_u8 *save_cdtv_dmac (int *len, uae_u8 *dstptr)
 {
 	uae_u8 *dstbak, *dst;
 	
-	if (!currprefs.cs_cdtvcd)
+	if (!currprefs.cs_cdtvcd || currprefs.cs_cdtvcr)
 		return NULL;
 	if (dstptr)
 		dstbak = dst = dstptr;
@@ -1798,7 +1732,7 @@ uae_u8 *save_dmac (int *len, uae_u8 *dstptr)
 
 }
 
-uae_u8 *restore_dmac (uae_u8 *src)
+uae_u8 *restore_cdtv_dmac (uae_u8 *src)
 {
 	restore_u32 ();
 	restore_u32 ();
@@ -1816,7 +1750,7 @@ uae_u8 *save_cdtv (int *len, uae_u8 *dstptr)
 {
 	uae_u8 *dstbak, *dst;
 
-	if (!currprefs.cs_cdtvcd)
+	if (!currprefs.cs_cdtvcd || currprefs.cs_cdtvcr)
 		return NULL;
 
 	if (dstptr) 
@@ -1851,7 +1785,7 @@ uae_u8 *save_cdtv (int *len, uae_u8 *dstptr)
 	save_u32 (last_play_pos);
 	save_u32 (last_play_end);
 	save_u64 (dma_wait);
-	for (int i = 0; i < sizeof cdrom_command_input; i++)
+	for (int i = 0; i < sizeof (cdrom_command_input); i++)
 		save_u8 (cdrom_command_input[i]);
 	save_u8 (cdrom_command_cnt_in);
 	save_u16 (cdtv_sectorsize);
@@ -1866,7 +1800,7 @@ uae_u8 *restore_cdtv (uae_u8 *src)
 	if (!currprefs.cs_cdtvcd) {
 		changed_prefs.cs_cdtvcd = changed_prefs.cs_cdtvram = true;
 		currprefs.cs_cdtvcd = currprefs.cs_cdtvram = true;
-		cdtv_init ();
+		cdtv_init (0);
 	}
 	restore_u32 ();
 	
@@ -1902,7 +1836,7 @@ uae_u8 *restore_cdtv (uae_u8 *src)
 	last_play_pos = restore_u32 ();
 	last_play_end = restore_u32 ();
 	dma_wait = restore_u64 ();
-	for (int i = 0; i < sizeof cdrom_command_input; i++)
+	for (int i = 0; i < sizeof (cdrom_command_input); i++)
 		cdrom_command_input[i] = restore_u8 ();
 	cdrom_command_cnt_in = restore_u8 ();
 	cdtv_sectorsize = restore_u16 ();
@@ -1912,9 +1846,9 @@ uae_u8 *restore_cdtv (uae_u8 *src)
 
 void restore_cdtv_finish (void)
 {
-	if (!currprefs.cs_cdtvcd)
+	if (!currprefs.cs_cdtvcd || currprefs.cs_cdtvcr)
 		return;
-	cdtv_init ();
+	cdtv_init (0);
 	get_toc ();
 	write_comm_pipe_u32 (&requests, 0x0104, 1);
 	if (cd_playing) {
