@@ -1,3 +1,7 @@
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
 #include <stdlib.h>
 #include <string.h>
 #include <uae/uae.h>
@@ -6,6 +10,7 @@
 #include <fs/i18n.h>
 #include <fs/time.h>
 #include "fs-uae.h"
+#include "config-common.h"
 
 #define NUM_SAVE_SLOTS 9
 /// TRANSLATORS: This is a menu entry and must not be too long: If you struggle
@@ -15,8 +20,8 @@
 /// to get in short enough, just translate "No Device" instead.
 #define NO_AMIGA_DEVICE _("No Amiga Device")
 #define PAUSE_ITEM_INDEX 1
-#define INPUT_ITEM_INDEX 8
-#define MEDIA_ITEM_INDEX 12
+#define INPUT_ITEM_INDEX 7
+#define MEDIA_ITEM_INDEX 11
 
 static int pause_function(fs_emu_menu_item *item, void **data) {
     fs_emu_log("pause_function\n");
@@ -67,9 +72,9 @@ static int reset_menu_function(fs_emu_menu_item *menu_item,
 
 static char *get_floppy_label(const char* path) {
     if (!path || path[0] == '\0') {
-        return fs_strdup("");
+        return g_strdup("");
     }
-    char *name = fs_path_get_basename(path);
+    char *name = g_path_get_basename(path);
 #ifdef USE_GLIB
     GError *error = NULL;
     GRegex *re = g_regex_new(
@@ -93,7 +98,7 @@ static char *get_floppy_label(const char* path) {
     if (!result) {
         return name;
     }
-    free(name);
+    g_free(name);
     return result;
 #else
     return name;
@@ -180,8 +185,8 @@ static void update_main_menu(fs_emu_menu *menu) {
     else {
         fs_emu_menu_item_set_title(item, _("Pause"));
     }
-    update_input_item(menu->items[INPUT_ITEM_INDEX], 0);
-    update_input_item(menu->items[INPUT_ITEM_INDEX + 1], 1);
+    update_input_item(menu->items[INPUT_ITEM_INDEX], 1);
+    update_input_item(menu->items[INPUT_ITEM_INDEX + 1], 0);
     int media_item_first_index = MEDIA_ITEM_INDEX;
 
     int drive, type;
@@ -200,7 +205,7 @@ static void update_main_menu(fs_emu_menu *menu) {
                 fs_emu_log("floppy in %d: %s\n", drive, path);
                 if (path == NULL || *path == 0) {
                     /// TRANSLATORS: Empty or nothing
-                    char *title = fs_strdup_printf(_("DF%d: Empty"), drive);
+                    char *title = g_strdup_printf(_("DF%d: Empty"), drive);
                     fs_emu_menu_item_set_title(item, title);
                     free(title);
                 }
@@ -217,7 +222,7 @@ static void update_main_menu(fs_emu_menu *menu) {
             fs_emu_log("CD in %d: %s\n", drive, path);
             if (path == NULL || *path == 0) {
                 /// TRANSLATORS: Empty or nothing
-                char *title = fs_strdup_printf(_("CD%d: Empty"), drive);
+                char *title = g_strdup_printf(_("CD%d: Empty"), drive);
                 fs_emu_menu_item_set_title(item, title);
                 free(title);
             }
@@ -259,7 +264,7 @@ static int save_function(fs_emu_menu_item *item, void **data) {
 }
 
 static char *get_state_file(int slot, const char *ext) {
-    char *path = fs_strdup_printf("%s %d.%s", fs_uae_get_state_base_name(),
+    char *path = g_strdup_printf("%s %d.%s", fs_uae_get_state_base_name(),
             slot + 1, ext);
     return path;
 }
@@ -281,13 +286,21 @@ static char *check_save_state(int slot) {
             //g_date_strftime(strbuf, 32, "");
             strftime(strbuf, 32, "%Y-%m-%d %H:%M:%S", &tm_struct);
             //title = fs_strdup_printf("%d", buf.mtime);
-            title = fs_strdup(strbuf);
+            title = g_strdup(strbuf);
         }
     }
-    free(state_file);
+    g_free(state_file);
     return title;
 }
 
+static int check_save_state_exists(int slot) {
+    char *state_file = get_state_file(slot, "uss");
+    fs_emu_log("check %s\n", state_file);
+    int result = fs_path_exists(state_file);
+    free(state_file);
+    return result;
+}
+
 static void update_save_state_item(fs_emu_menu_item* item, int slot,
         int save) {
     char *title = check_save_state(slot);
@@ -301,13 +314,13 @@ static void update_save_state_item(fs_emu_menu_item* item, int slot,
 }
 
 static void update_save_states_menu(fs_emu_menu *menu) {
-    fs_emu_menu_item *item;
     for (int i = 0; i < NUM_SAVE_SLOTS; i++) {
-        item = menu->items[1 + i];
+        fs_emu_menu_item *item = menu->items[1 + i];
         update_save_state_item(item, i, 1);
     }
 }
 
+#if 0
 static int save_state_menu_function(fs_emu_menu_item *menu_item,
         void **result_data) {
 
@@ -348,6 +361,37 @@ static int save_state_menu_function(fs_emu_menu_item *menu_item,
     *result_data = menu;
     return FS_EMU_MENU_RESULT_MENU;
 }
+#endif
+
+static int load_states_menu_function(fs_emu_menu_item *unused,
+        void **result_data) {
+    fs_emu_log("load_states_menu_function\n");
+    fs_emu_menu_item *item;
+    fs_emu_menu *menu = fs_emu_menu_new();
+    fs_emu_menu_set_update_function(menu, update_save_states_menu);
+
+    item = fs_emu_menu_item_new();
+    fs_emu_menu_append_item(menu, item);
+    fs_emu_menu_item_set_title(item, _("Load State"));
+    fs_emu_menu_item_set_type(item, FS_EMU_MENU_ITEM_TYPE_HEADING);
+
+    for (int i = 0; i < NUM_SAVE_SLOTS; i++) {
+        item = fs_emu_menu_item_new();
+        fs_emu_menu_append_item(menu, item);
+        fs_emu_menu_item_set_idata(item, i);
+        fs_emu_menu_item_set_enabled(item, check_save_state_exists(i));
+        fs_emu_menu_item_set_activate_function(item,
+                //save_state_menu_function);
+                load_function);
+    }
+
+    // focus on the last used slot
+    menu->index = g_last_save_slot + 1;
+
+    //create_save_state_menu(menu, 1);
+    *result_data = menu;
+    return FS_EMU_MENU_RESULT_MENU;
+}
 
 static int save_states_menu_function(fs_emu_menu_item *unused,
         void **result_data) {
@@ -358,7 +402,7 @@ static int save_states_menu_function(fs_emu_menu_item *unused,
 
     item = fs_emu_menu_item_new();
     fs_emu_menu_append_item(menu, item);
-    fs_emu_menu_item_set_title(item, _("Save States"));
+    fs_emu_menu_item_set_title(item, _("Save State"));
     fs_emu_menu_item_set_type(item, FS_EMU_MENU_ITEM_TYPE_HEADING);
 
     for (int i = 0; i < NUM_SAVE_SLOTS; i++) {
@@ -366,7 +410,8 @@ static int save_states_menu_function(fs_emu_menu_item *unused,
         fs_emu_menu_append_item(menu, item);
         fs_emu_menu_item_set_idata(item, i);
         fs_emu_menu_item_set_activate_function(item,
-                save_state_menu_function);
+                //save_state_menu_function);
+                save_function);
     }
 
     // focus on the last used slot
@@ -400,13 +445,14 @@ static void insert_cdrom(int drive_index, int disk_index) {
     }
     fs_emu_log("menu: insert CD index %d into drive %d\n", disk_index,
             drive_index);
-    char *key = fs_strdup_printf("cdrom_image_%d", disk_index);
+    char *key = g_strdup_printf("cdrom_image_%d", disk_index);
     char* path = fs_config_get_string(key);
     free(key);
     if (path == NULL) {
         fs_emu_log("no CD at this index in CD-ROM list\n");
     }
     path = fs_uae_expand_path_and_free(path);
+    path = fs_uae_resolve_path_and_free(path, FS_UAE_CD_PATHS);
     amiga_cdrom_set_file(drive_index, path);
     free(path);
 }
@@ -475,7 +521,7 @@ static void update_cd_menu(fs_emu_menu *menu) {
     //fs_emu_menu_item_set_enabled(item, inserted_path[0] != '\0');
     for (int i = 0; i < menu->count - 2; i++) {
         item = menu->items[i + 2];
-        char *key = fs_strdup_printf("cdrom_image_%d", i);
+        char *key = g_strdup_printf("cdrom_image_%d", i);
         char* path = fs_config_get_string(key);
         free(key);
         if (path == NULL) {
@@ -511,7 +557,7 @@ static int media_menu_function(fs_emu_menu_item *menu_item,
         item = fs_emu_menu_item_new();
         fs_emu_menu_append_item(menu, item);
         /// TRANSLATORS: This is a menu entry and must not be too long
-        str = fs_strdup_printf(_("Insert Into DF%d"), drive);
+        str = g_strdup_printf(_("Insert Into DF%d"), drive);
         fs_emu_menu_item_set_title(item, str);
         free(str);
         fs_emu_menu_item_set_type(item, FS_EMU_MENU_ITEM_TYPE_HEADING);
@@ -553,9 +599,9 @@ static int media_menu_function(fs_emu_menu_item *menu_item,
         item = fs_emu_menu_item_new();
         fs_emu_menu_append_item(menu, item);
         /// TRANSLATORS: This is a menu entry and must not be too long
-        str = fs_strdup_printf(_("Insert Into CD%d"), drive);
+        str = g_strdup_printf(_("Insert Into CD%d"), drive);
         fs_emu_menu_item_set_title(item, str);
-        free(str);
+        g_free(str);
         fs_emu_menu_item_set_type(item, FS_EMU_MENU_ITEM_TYPE_HEADING);
 
         item = fs_emu_menu_item_new();
@@ -569,7 +615,7 @@ static int media_menu_function(fs_emu_menu_item *menu_item,
 
         for (int i = 0; ; i++) {
             // FIXME: GET FROM UAE OPTIONS...?
-            char *key = fs_strdup_printf("cdrom_image_%d", i);
+            char *key = g_strdup_printf("cdrom_image_%d", i);
             char *path = fs_config_get_string(key);
             free(key);
             if (path == NULL) {
@@ -664,6 +710,7 @@ static int input_device_function(fs_emu_menu_item *menu_item,
         fs_log("[menu] port %d set device to \"%s\"\n", port, "");
         set_input_port(port, "", 0);
     }
+#if 0
     else if (index == 254) {
         fs_log("[menu] port %d set device to \"%s\"\n", port, "MOUSE");
         set_input_port(port, "MOUSE", 1);
@@ -672,6 +719,7 @@ static int input_device_function(fs_emu_menu_item *menu_item,
         fs_log("[menu] port %d set device to \"%s\"\n", port, "KEYBOARD");
         set_input_port(port, "KEYBOARD", 1);
     }
+#endif
     else {
         fs_emu_input_device device;
         if (!fs_ml_input_device_get(index, &device)) {
@@ -695,9 +743,9 @@ static int input_host_menu_function(fs_emu_menu_item *menu_item,
     item = fs_emu_menu_item_new();
     fs_emu_menu_append_item(menu, item);
     /// TRANSLATORS: This is a menu entry and must not be too long
-    str = fs_strdup_printf(_("Joystick Port %d Device"), port);
+    str = g_strdup_printf(_("Joystick Port %d Device"), port);
     fs_emu_menu_item_set_title(item, str);
-    free(str);
+    g_free(str);
     fs_emu_menu_item_set_type(item, FS_EMU_MENU_ITEM_TYPE_HEADING);
 
     item = fs_emu_menu_item_new();
@@ -709,6 +757,7 @@ static int input_host_menu_function(fs_emu_menu_item *menu_item,
         menu->index = 1;
     }
 
+#if 0
     item = fs_emu_menu_item_new();
     fs_emu_menu_append_item(menu, item);
     fs_emu_menu_item_set_title(item, _("Mouse"));
@@ -727,22 +776,32 @@ static int input_host_menu_function(fs_emu_menu_item *menu_item,
             "KEYBOARD") == 0) {
         menu->index = 3;
     }
+#endif
 
     fs_emu_input_device device;
     for (int i = 0; i < FS_ML_INPUT_DEVICES_MAX; i++) {
         if (!fs_ml_input_device_get(i, &device)) {
             continue;
         }
+        /*
         if (strcmp(device.name, "KEYBOARD") == 0) {
             continue;
         }
+        */
         if (strcmp(g_fs_uae_input_ports[port].device,
                 device.name) == 0) {
-            menu->index = 4 + i;
+            // menu->index = 4 + i;
+            menu->index = 2 + i;
         }
         item = fs_emu_menu_item_new();
         fs_emu_menu_append_item(menu, item);
         const char* s = device.name;
+        if (strcmp(s, "KEYBOARD") == 0) {
+            s = _("Keyboard");
+        }
+        else if (strcmp(s, "MOUSE") == 0) {
+            s = _("Mouse");
+        }
         fs_emu_menu_item_set_title(item, s);
         fs_emu_menu_item_set_idata(item, (port << 8) | i);
         fs_emu_menu_item_set_activate_function(item, input_device_function);
@@ -782,9 +841,9 @@ static int input_amiga_menu_function(fs_emu_menu_item *menu_item,
     item = fs_emu_menu_item_new();
     fs_emu_menu_append_item(menu, item);
     /// TRANSLATORS: This is a menu entry and must not be too long
-    str = fs_strdup_printf(_("Joystick Port %d Mode"), port);
+    str = g_strdup_printf(_("Joystick Port %d Mode"), port);
     fs_emu_menu_item_set_title(item, str);
-    free(str);
+    g_free(str);
     fs_emu_menu_item_set_type(item, FS_EMU_MENU_ITEM_TYPE_HEADING);
 
     int index = 0;
@@ -876,9 +935,17 @@ static int input_menu_function(fs_emu_menu_item *menu_item,
     item = fs_emu_menu_item_new();
     fs_emu_menu_append_item(menu, item);
     /// TRANSLATORS: This is a menu entry and must not be too long
-    str = fs_strdup_printf(_("Joystick Port %d"), port);
+    if (port == 0) {
+        str = g_strdup_printf(_("Mouse Port"));
+    }
+    else if (port == 1) {
+        str = g_strdup_printf(_("Joystick Port"));
+    }
+    else {
+        str = g_strdup_printf(_("Joystick Port %d"), port);
+    }
     fs_emu_menu_item_set_title(item, str);
-    free(str);
+    g_free(str);
     fs_emu_menu_item_set_type(item, FS_EMU_MENU_ITEM_TYPE_HEADING);
 
     item = fs_emu_menu_item_new();
@@ -910,16 +977,17 @@ static int input_menu_function(fs_emu_menu_item *menu_item,
     return FS_EMU_MENU_RESULT_MENU;
 }
 
-void add_input_item(fs_emu_menu *menu, int index) {
+static void add_input_item(fs_emu_menu *menu, int index) {
     fs_emu_menu_item *item;
 
     item = fs_emu_menu_item_new();
     fs_emu_menu_append_item(menu, item);
     if (index == 0) {
-        fs_emu_menu_item_set_title(item, _("Joystick Port 0"));
+        fs_emu_menu_item_set_title(item, _("Mouse Port"));
     }
     else if (index == 1) {
-        fs_emu_menu_item_set_title(item, _("Joystick Port 1"));
+
+        fs_emu_menu_item_set_title(item, _("Joystick Port"));
     }
     else if (index == 2) {
         fs_emu_menu_item_set_title(item, _("Joystick Port 2"));
@@ -939,11 +1007,11 @@ void add_input_item(fs_emu_menu *menu, int index) {
 }
 
 static void update_input_options_menu(fs_emu_menu *menu) {
-    update_input_item(menu->items[1], 0);
-    update_input_item(menu->items[2], 1);
-    update_input_item(menu->items[4], 2);
-    update_input_item(menu->items[5], 3);
-    update_input_item(menu->items[7], 4);
+    update_input_item(menu->items[1], 1);
+    update_input_item(menu->items[3], 0);
+    update_input_item(menu->items[5], 2);
+    update_input_item(menu->items[6], 3);
+    update_input_item(menu->items[8], 4);
 }
 
 static int input_options_menu_function(fs_emu_menu_item *menu_item,
@@ -956,11 +1024,17 @@ static int input_options_menu_function(fs_emu_menu_item *menu_item,
 
     item = fs_emu_menu_item_new();
     fs_emu_menu_append_item(menu, item);
-    fs_emu_menu_item_set_title(item, _("Joystick Ports"));
+    fs_emu_menu_item_set_title(item, _("Joystick Port"));
+    fs_emu_menu_item_set_type(item, FS_EMU_MENU_ITEM_TYPE_HEADING);
+
+    add_input_item(menu, 1);
+
+    item = fs_emu_menu_item_new();
+    fs_emu_menu_append_item(menu, item);
+    fs_emu_menu_item_set_title(item, _("Mouse Port"));
     fs_emu_menu_item_set_type(item, FS_EMU_MENU_ITEM_TYPE_HEADING);
 
     add_input_item(menu, 0);
-    add_input_item(menu, 1);
 
     item = fs_emu_menu_item_new();
     fs_emu_menu_append_item(menu, item);
@@ -1001,18 +1075,26 @@ void fs_uae_configure_menu() {
 
     item = fs_emu_menu_item_new();
     fs_emu_menu_append_item(menu, item);
-    fs_emu_menu_item_set_title(item, _("Save States"));
-    fs_emu_menu_item_set_activate_function(item, save_states_menu_function);
+    fs_emu_menu_item_set_title(item, _("Load State"));
+    fs_emu_menu_item_set_activate_function(item, load_states_menu_function);
     if (fs_config_get_boolean("save_states") == 0) {
         fs_emu_menu_item_set_enabled(item, 0);
     }
 
+    item = fs_emu_menu_item_new();
+    fs_emu_menu_append_item(menu, item);
+    fs_emu_menu_item_set_title(item, _("Save State"));
+    fs_emu_menu_item_set_activate_function(item, save_states_menu_function);
+    if (fs_config_get_boolean("save_states") == 0) {
+        fs_emu_menu_item_set_enabled(item, 0);
+    }
+/*
     item = fs_emu_menu_item_new();
     fs_emu_menu_append_item(menu, item);
     fs_emu_menu_item_set_title(item, _("More..."));
     fs_emu_menu_item_set_type(item, FS_EMU_MENU_ITEM_TYPE_MENU);
     fs_emu_menu_item_set_enabled(item, 0);
-
+*/
     item = fs_emu_menu_item_new();
     fs_emu_menu_append_item(menu, item);
     fs_emu_menu_item_set_title(item, _("Amiga Control"));
@@ -1022,20 +1104,20 @@ void fs_uae_configure_menu() {
     fs_emu_menu_append_item(menu, item);
     fs_emu_menu_item_set_title(item, _("Reset Amiga"));
     fs_emu_menu_item_set_activate_function(item, reset_menu_function);
-
+/*
     item = fs_emu_menu_item_new();
     fs_emu_menu_append_item(menu, item);
     fs_emu_menu_item_set_title(item, _("More..."));
     fs_emu_menu_item_set_type(item, FS_EMU_MENU_ITEM_TYPE_MENU);
     fs_emu_menu_item_set_enabled(item, 0);
-
+*/
     item = fs_emu_menu_item_new();
     fs_emu_menu_append_item(menu, item);
     fs_emu_menu_item_set_title(item, _("Input Options"));
     fs_emu_menu_item_set_type(item, FS_EMU_MENU_ITEM_TYPE_HEADING);
 
-    add_input_item(menu, 0);
     add_input_item(menu, 1);
+    add_input_item(menu, 0);
 
     item = fs_emu_menu_item_new();
     fs_emu_menu_append_item(menu, item);
