@@ -1,9 +1,16 @@
+#ifndef UAE_CDTV_H
+#define UAE_CDTV_H
+
+#ifdef FSUAE // NL
+#include "uae/types.h"
+#include "uae/memory.h"
+#endif
 
 #ifdef CDTV
 
 extern addrbank dmac_bank;
 
-extern void cdtv_init (void);
+extern addrbank *cdtv_init (struct romconfig *rc);
 extern void cdtv_free (void);
 extern void CDTV_hsync_handler(void);
 extern void cdtv_check_banks (void);
@@ -14,9 +21,8 @@ uae_u8 cdtv_battram_read (int addr);
 extern void cdtv_loadcardmem (uae_u8*, int);
 extern void cdtv_savecardmem (uae_u8*, int);
 
-int cdtv_add_scsi_unit (int ch, TCHAR *path, int blocksize, int readonly,
-		       TCHAR *devname, int sectors, int surfaces, int reserved,
-		       int bootpri, TCHAR *filesys);
+extern void cdtv_add_scsi_unit (int ch, struct uaedev_config_info *ci, struct romconfig *rc);
+
 
 extern void cdtv_getdmadata (uae_u32*);
 
@@ -26,5 +32,6 @@ extern void cdtv_scsi_clear_int (void);
 
 extern bool cdtv_front_panel (int);
 
-#endif
+#endif // CDTV
 
+#endif // UAE_CDTV_H
