@@ -1,3 +1,10 @@
+#ifndef UAE_DONGLE_H
+#define UAE_DONGLE_H
+
+#ifdef FSUAE // NL
+#include "uae/types.h"
+#endif
+
 extern void dongle_reset (void);
 extern uae_u8 dongle_cia_read (int, int, uae_u8);
 extern void dongle_cia_write (int, int, uae_u8);
@@ -5,4 +12,6 @@ extern void dongle_joytest (uae_u16);
 extern uae_u16 dongle_joydat (int, uae_u16);
 extern void dongle_potgo (uae_u16);
 extern uae_u16 dongle_potgor (uae_u16);
-extern int dongle_analogjoy (int, int);
\ No newline at end of file
+extern int dongle_analogjoy (int, int);
+
+#endif // UAE_DONGLE_H
