@@ -6,6 +6,12 @@
   * Copyright 2004 James Bagg, Toni Wilen
   */
 
+#ifndef UAE_DRIVECLICK_H
+#define UAE_DRIVECLICK_H
+
+#ifdef FSUAE // NL
+#include "uae/types.h"
+#endif
 
 #define CLICK_TRACKS 84
 
@@ -48,4 +54,6 @@ extern void driveclick_fdrawcmd_detect (void);
 extern void driveclick_fdrawcmd_seek (int, int);
 extern void driveclick_fdrawcmd_motor (int, int);
 extern void driveclick_fdrawcmd_vsync (void);
-extern int driveclick_pcdrivemask, driveclick_pcdrivenum;
\ No newline at end of file
+extern int driveclick_pcdrivemask, driveclick_pcdrivenum;
+
+#endif // UAE_DRIVECLICK_H
