@@ -6,7 +6,16 @@
   * (c) 1995 Bernd Schmidt
   */
 
+#ifndef UAE_ERSATZ_H
+#define UAE_ERSATZ_H
+
+#ifdef FSUAE // NL
+#include "uae/types.h"
+#endif
+
 extern void init_ersatz_rom (uae_u8 *data);
 extern void ersatz_chipcopy (void);
 extern void ersatz_perform (uae_u16);
 extern void DISK_ersatz_read (int,int, uaecptr);
+
+#endif // UAE_ERSATZ_H
