@@ -6,6 +6,9 @@
   * Copyright 1996 Bernd Schmidt
   */
 
+#ifndef UAE_EXECLIB_H
+#define UAE_EXECLIB_H
+
 #define CMD_INVALID 0
 #define CMD_RESET   1
 #define CMD_READ    2
@@ -45,3 +48,5 @@
 #define MEMF_REVERSE (1<<18)
 #define MEMF_TOTAL (1<<19)
 #endif
+
+#endif // UAE_EXECLIB_H
