@@ -7,9 +7,14 @@
   * Copyright 1996 Alessandro Bissacco
   */
 
+#ifndef UAE_GENBLITTER_H
+#define UAE_GENBLITTER_H
+
 struct blitop {
     const char *s;
     int used;
 };
 
 extern struct blitop blitops[256];
+
+#endif // UAE_GENBLITTER_H
