@@ -7,6 +7,9 @@
   * Copyright 1997 Bernd Schmidt
   */
 
+#ifndef UAE_GENSOUND_H
+#define UAE_GENSOUND_H
+
 extern int sound_available;
 
 extern void (*sample_handler) (void);
@@ -27,3 +30,4 @@ extern void sample16s_handler (void);
 extern void sample16ss_handler (void);
 extern void sample8s_handler (void);
 
+#endif // UAE_GENSOUND_H
