@@ -1,3 +1,11 @@
+#ifndef UAE_ISOFS_API_H
+#define UAE_ISOFS_API_H
+
+#ifdef FSUAE // NL
+#include "config.h"
+#include "uae/types.h"
+#include <time.h>
+#endif
 
 struct cd_opendir_s;
 struct cd_openfile_s;
@@ -30,3 +38,5 @@ void isofs_closefile(struct cd_openfile_s*);
 uae_s64 isofs_lseek(struct cd_openfile_s*, uae_s64, int);
 uae_s64 isofs_fsize(struct cd_openfile_s*);
 uae_s64 isofs_read(struct cd_openfile_s*, void*, unsigned int);
+
+#endif // UAE_ISOFS_API_H
