@@ -3,6 +3,14 @@
  * Amiga counterpart, but can be called in situation where calling
  * the exec.library functions is impossible.
  */
+
+#ifndef UAE_NATIVE2AMIGA_API_H
+#define UAE_NATIVE2AMIGA_API_H
+
+#ifdef FSUAE
+#include "uae/types.h"
+#endif
+
 #ifdef SUPPORT_THREADS
 void uae_Cause(uaecptr interrupt);
 void uae_ReplyMsg(uaecptr msg);
@@ -11,4 +19,6 @@ void uae_Signal(uaecptr task, uae_u32 mask);
 void uae_NotificationHack(uaecptr, uaecptr);
 #endif
 void uae_NewList(uaecptr list);
-int native2amiga_isfree (void);
\ No newline at end of file
+int native2amiga_isfree (void);
+
+#endif // UAE_NATIVE2AMIGA_API_H
