@@ -7,9 +7,17 @@
 * Copyright 1995-2001 Bernd Schmidt
 */
 
-#define UAEMAJOR 2
-#define UAEMINOR 5
-#define UAESUBREV 1
+#ifndef UAE_OPTIONS_H
+#define UAE_OPTIONS_H
+
+#ifdef FSUAE // NL
+#include "uae/types.h"
+#include "uae/limits.h"
+#endif
+
+#define UAEMAJOR 3
+#define UAEMINOR 1
+#define UAESUBREV 0
 
 typedef enum { KBD_LANG_US, KBD_LANG_DK, KBD_LANG_DE, KBD_LANG_SE, KBD_LANG_FR, KBD_LANG_IT, KBD_LANG_ES } KbdLang;
 
@@ -30,7 +38,7 @@ struct strlist {
 #define MAX_TOTAL_SCSI_DEVICES 8
 
 /* maximum number native input devices supported (single type) */
-#define MAX_INPUT_DEVICES 16
+#define MAX_INPUT_DEVICES 20
 /* maximum number of native input device's buttons and axles supported */
 #define MAX_INPUT_DEVICE_EVENTS 256
 /* 4 different customization settings */
@@ -63,6 +71,7 @@ struct jport {
 	int autofire;
 	TCHAR name[MAX_JPORTNAME];
 	TCHAR configname[MAX_JPORTNAME];
+	bool nokeyboardoverride;
 };
 #define JPORT_NONE -1
 #define JPORT_CUSTOM -2
@@ -77,19 +86,34 @@ struct jport {
 #define MAX_SPARE_DRIVES 20
 #define MAX_CUSTOM_MEMORY_ADDRS 2
 
+#define CONFIG_TYPE_ALL -1
+#define CONFIG_TYPE_DEFAULT 0
 #define CONFIG_TYPE_HARDWARE 1
 #define CONFIG_TYPE_HOST 2
+#define CONFIG_TYPE_NORESET 4
 #define CONFIG_BLEN 2560
 
 #define TABLET_OFF 0
 #define TABLET_MOUSEHACK 1
 #define TABLET_REAL 2
 
+#ifdef WITH_SLIRP
+#define MAX_SLIRP_REDIRS 32
+struct slirp_redir
+{
+	int proto;
+	int srcport;
+	int dstport;
+	unsigned long addr;
+};
+#endif
+
 struct cdslot
 {
 	TCHAR name[MAX_DPATH];
 	bool inuse;
 	bool delayed;
+	bool temporary;
 	int type;
 };
 struct floppyslot
@@ -98,37 +122,74 @@ struct floppyslot
 	int dfxtype;
 	int dfxclick;
 	TCHAR dfxclickexternal[256];
+	bool forcedwriteprotect;
 };
 
+#define ASPECTMULT 1024
+#define WH_NATIVE 1
 struct wh {
 	int x, y;
 	int width, height;
+	int special;
 };
 
 #define MOUNT_CONFIG_SIZE 30
+#define UAEDEV_DIR 0
+#define UAEDEV_HDF 1
+#define UAEDEV_CD 2
+#define UAEDEV_TAPE 3
+
+#define BOOTPRI_NOAUTOBOOT -128
+#define BOOTPRI_NOAUTOMOUNT -129
+#define ISAUTOBOOT(ci) ((ci)->bootpri > BOOTPRI_NOAUTOBOOT)
+#define ISAUTOMOUNT(ci) ((ci)->bootpri > BOOTPRI_NOAUTOMOUNT)
 struct uaedev_config_info {
+	int type;
 	TCHAR devname[MAX_DPATH];
 	TCHAR volname[MAX_DPATH];
 	TCHAR rootdir[MAX_DPATH];
-	bool ishdf;
 	bool readonly;
 	int bootpri;
-	bool autoboot;
-	bool donotmount;
 	TCHAR filesys[MAX_DPATH];
-	int cyls; // zero if detected from size
+	int lowcyl;
+	int highcyl; // zero if detected from size
+	int cyls; // calculated/corrected highcyl
 	int surfaces;
 	int sectors;
 	int reserved;
 	int blocksize;
-	int configoffset;
-	int controller;
-	// zero if default
+	int controller_type;
+	int controller_type_unit;
+	int controller_unit;
+	int controller_media_type; // 1 = CF IDE, 0 = normal
+	int unit_feature_level;
+	bool physical_geometry; // if false: use defaults
 	int pcyls, pheads, psecs;
+	int flags;
+	int buffers;
+	int bufmemtype;
+	int stacksize;
+	int priority;
+	uae_u32 mask;
+	int maxtransfer;
+	uae_u32 dostype;
+	int unit;
+	int interleave;
+	int sectorsperblock;
+	int forceload;
+	int device_emu_unit;
+	bool inject_icons;
 };
 
-enum { CP_GENERIC = 1, CP_CDTV, CP_CD32, CP_A500, CP_A500P, CP_A600, CP_A1000,
-	CP_A1200, CP_A2000, CP_A3000, CP_A3000T, CP_A4000, CP_A4000T };
+struct uaedev_config_data
+{
+	struct uaedev_config_info ci;
+	int configoffset; // HD config entry index
+	int unitnum; // scsi unit number (if tape currently)
+};
+
+enum { CP_GENERIC = 1, CP_CDTV, CP_CDTVCR, CP_CD32, CP_A500, CP_A500P, CP_A600, CP_A1000,
+	CP_A1200, CP_A2000, CP_A3000, CP_A3000T, CP_A4000, CP_A4000T, CP_VELVET };
 
 #define IDE_A600A1200 1
 #define IDE_A4000 2
@@ -147,11 +208,22 @@ enum { CP_GENERIC = 1, CP_CDTV, CP_CD32, CP_A500, CP_A500P, CP_A600, CP_A1000,
 #define AUTOSCALE_MANUAL 7 // use gfx_xcenter_pos and gfx_ycenter_pos
 #define AUTOSCALE_INTEGER 8
 #define AUTOSCALE_INTEGER_AUTOSCALE 9
+#define AUTOSCALE_SEPARATOR 10
+#define AUTOSCALE_OVERSCAN_BLANK 11
 
 #define MONITOREMU_NONE 0
 #define MONITOREMU_AUTO 1
 #define MONITOREMU_A2024 2
 #define MONITOREMU_GRAFFITI 3
+#define MONITOREMU_HAM_E 4
+#define MONITOREMU_HAM_E_PLUS 5
+#define MONITOREMU_VIDEODAC18 6
+#define MONITOREMU_AVIDEO12 7
+#define MONITOREMU_AVIDEO24 8
+#define MONITOREMU_FIRECRACKER24 9
+#define MONITOREMU_DCTV 10
+
+#define MAX_FILTERSHADERS 4
 
 #define MAX_CHIPSET_REFRESH 10
 #define MAX_CHIPSET_REFRESH_TOTAL (MAX_CHIPSET_REFRESH + 2)
@@ -185,12 +257,69 @@ struct apmode
 	// -1 = wait for flip, before frame ends
 	// 1 = wait for flip, after new frame has started
 	int gfx_vflip;
+	// doubleframemode strobo
+	bool gfx_strobo;
 	int gfx_vsyncmode;
 	int gfx_backbuffers;
 	bool gfx_interlaced;
 	int gfx_refreshrate;
 };
 
+#define MAX_LUA_STATES 16
+
+
+struct gfx_filterdata
+{
+	int gfx_filter;
+	TCHAR gfx_filtershader[2 * MAX_FILTERSHADERS + 1][MAX_DPATH];
+	TCHAR gfx_filtermask[2 * MAX_FILTERSHADERS + 1][MAX_DPATH];
+	TCHAR gfx_filteroverlay[MAX_DPATH];
+	struct wh gfx_filteroverlay_pos;
+	int gfx_filteroverlay_overscan;
+	int gfx_filter_scanlines;
+	int gfx_filter_scanlineratio;
+	int gfx_filter_scanlinelevel;
+	float gfx_filter_horiz_zoom, gfx_filter_vert_zoom;
+	float gfx_filter_horiz_zoom_mult, gfx_filter_vert_zoom_mult;
+	float gfx_filter_horiz_offset, gfx_filter_vert_offset;
+	int gfx_filter_left_border, gfx_filter_right_border;
+	int gfx_filter_top_border, gfx_filter_bottom_border;
+	int gfx_filter_filtermode;
+	int gfx_filter_bilinear;
+	int gfx_filter_noise, gfx_filter_blur;
+	int gfx_filter_saturation, gfx_filter_luminance, gfx_filter_contrast;
+	int gfx_filter_gamma, gfx_filter_gamma_ch[3];
+	int gfx_filter_keep_aspect, gfx_filter_aspect;
+	int gfx_filter_autoscale;
+	int gfx_filter_integerscalelimit;
+	int gfx_filter_keep_autoscale_aspect;
+};
+
+#define MAX_DUPLICATE_EXPANSION_BOARDS 4
+#define MAX_EXPANSION_BOARDS 4
+struct romconfig
+{
+	TCHAR romfile[MAX_DPATH];
+	TCHAR romident[256];
+	uae_u32 board_ram_size;
+	bool autoboot_disabled;
+	int device_id;
+	int device_settings;
+	int subtype;
+	void *unitdata;
+};
+#define MAX_BOARD_ROMS 2
+struct boardromconfig
+{
+	int device_type;
+	int device_num;
+	struct romconfig roms[MAX_BOARD_ROMS];
+};
+
+#define Z3MAPPING_AUTO 0
+#define Z3MAPPING_UAE 1
+#define Z3MAPPING_REAL 2
+
 struct uae_prefs {
 
 	struct strlist *all_lines;
@@ -208,6 +337,7 @@ struct uae_prefs {
 	bool serial_hwctsrts;
 	bool serial_direct;
 	int serial_stopbits;
+	int serial_crlf;
 	bool parallel_demand;
 	int parallel_matrix_emulation;
 	bool parallel_postscript_emulation;
@@ -231,11 +361,14 @@ struct uae_prefs {
 	int sound_interpol;
 	int sound_filter;
 	int sound_filter_type;
-	int sound_volume;
+	int sound_volume_master;
+	int sound_volume_paula;
 	int sound_volume_cd;
+	int sound_volume_board;
 	bool sound_stereo_swap_paula;
 	bool sound_stereo_swap_ahi;
 	bool sound_auto;
+	bool sound_cdaudio;
 
 	int sampler_freq;
 	int sampler_buffer;
@@ -250,6 +383,7 @@ struct uae_prefs {
 	bool comp_midopt;
 	bool comp_lowopt;
 	bool fpu_strict;
+	bool fpu_softfloat;
 
 	bool comp_hardflush;
 	bool comp_constjump;
@@ -266,46 +400,31 @@ struct uae_prefs {
 	struct wh gfx_size;
 	struct wh gfx_size_win_xtra[6];
 	struct wh gfx_size_fs_xtra[6];
-	bool gfx_autoresolution;
+	bool gfx_autoresolution_vga;
+	int gfx_autoresolution;
+	int gfx_autoresolution_delay;
 	int gfx_autoresolution_minv, gfx_autoresolution_minh;
 	bool gfx_scandoubler;
 	struct apmode gfx_apmode[2];
 	int gfx_resolution;
 	int gfx_vresolution;
 	int gfx_lores_mode;
-	int gfx_scanlines;
+	int gfx_pscanlines, gfx_iscanlines;
 	int gfx_xcenter, gfx_ycenter;
 	int gfx_xcenter_pos, gfx_ycenter_pos;
 	int gfx_xcenter_size, gfx_ycenter_size;
 	int gfx_max_horizontal, gfx_max_vertical;
-	int gfx_saturation, gfx_luminance, gfx_contrast, gfx_gamma;
+	int gfx_saturation, gfx_luminance, gfx_contrast, gfx_gamma, gfx_gamma_ch[3];
 	bool gfx_blackerthanblack;
 	int gfx_api;
 	int color_mode;
 	int gfx_extrawidth;
+	bool lightboost_strobo;
 
-	int gfx_filter;
-	TCHAR gfx_filtershader[MAX_DPATH];
-	TCHAR gfx_filtermask[MAX_DPATH];
-	TCHAR gfx_filteroverlay[MAX_DPATH];
-	struct wh gfx_filteroverlay_pos;
-	int gfx_filteroverlay_overscan;
-	int gfx_filter_scanlines;
-	int gfx_filter_scanlineratio;
-	int gfx_filter_scanlinelevel;
-	int gfx_filter_horiz_zoom, gfx_filter_vert_zoom;
-	int gfx_filter_horiz_zoom_mult, gfx_filter_vert_zoom_mult;
-	int gfx_filter_horiz_offset, gfx_filter_vert_offset;
-	int gfx_filter_filtermode;
-	int gfx_filter_bilinear;
-	int gfx_filter_noise, gfx_filter_blur;
-	int gfx_filter_saturation, gfx_filter_luminance, gfx_filter_contrast, gfx_filter_gamma;
-	int gfx_filter_keep_aspect, gfx_filter_aspect;
-	int gfx_filter_autoscale;
-	int gfx_filter_keep_autoscale_aspect;
+	struct gfx_filterdata gf[2];
 
-	int rtg_horiz_zoom_mult;
-	int rtg_vert_zoom_mult;
+	float rtg_horiz_zoom_mult;
+	float rtg_vert_zoom_mult;
 
 	bool immediate_blits;
 	int waiting_blits;
@@ -327,6 +446,7 @@ struct uae_prefs {
 	bool uaeserial;
 	int catweasel;
 	int cpu_idle;
+	int ppc_cpu_idle;
 	bool cpu_cycle_exact;
 	int cpu_clock_multiplier;
 	int cpu_frequency;
@@ -336,11 +456,22 @@ struct uae_prefs {
 	int floppy_random_bits_min;
 	int floppy_random_bits_max;
 	int floppy_auto_ext2;
+	int cd_speed;
 	bool tod_hack;
 	uae_u32 maprom;
+	int boot_rom;
+	bool rom_readwrite;
 	int turbo_emulation;
 	bool headless;
 	int filesys_limit;
+	int filesys_max_name;
+	int filesys_max_file_size;
+	bool filesys_inject_icons;
+	TCHAR filesys_inject_icons_tool[MAX_DPATH];
+	TCHAR filesys_inject_icons_project[MAX_DPATH];
+	TCHAR filesys_inject_icons_drawer[MAX_DPATH];
+	int uaescsidevmode;
+	bool reset_delay;
 
 	int cs_compatible;
 	int cs_ciaatod;
@@ -353,6 +484,8 @@ struct uae_prefs {
 	bool cs_cd32cd;
 	bool cs_cd32c2p;
 	bool cs_cd32nvram;
+	bool cs_cd32fmv;
+	int cs_cd32nvram_size;
 	bool cs_cdtvcd;
 	bool cs_cdtvram;
 	int cs_cdtvcard;
@@ -365,13 +498,20 @@ struct uae_prefs {
 	int cs_deniserev;
 	int cs_mbdmac;
 	bool cs_cdtvscsi;
-	bool cs_a2091, cs_a4091;
+	bool cs_cdtvcr;
 	bool cs_df0idhw;
 	bool cs_slowmemisfast;
 	bool cs_resetwarning;
 	bool cs_denisenoehb;
 	bool cs_dipagnus;
 	bool cs_agnusbltbusybug;
+	bool cs_ciatodbug;
+	bool cs_z3autoconfig;
+	bool cs_1mchipjumper;
+	bool cs_cia6526;
+	int cs_hacks;
+
+	struct boardromconfig expansionboard[MAX_EXPANSION_BOARDS];
 
 	TCHAR romfile[MAX_DPATH];
 	TCHAR romident[256];
@@ -387,8 +527,8 @@ struct uae_prefs {
 	TCHAR pci_devices[256];
 	TCHAR prtname[256];
 	TCHAR sername[256];
-	TCHAR amaxromfile[MAX_DPATH];
 	TCHAR a2065name[MAX_DPATH];
+	TCHAR picassoivromfile[MAX_DPATH];
 	struct cdslot cdslots[MAX_TOTAL_SCSI_DEVICES];
 	TCHAR quitstatefile[MAX_DPATH];
 	TCHAR statefile[MAX_DPATH];
@@ -407,13 +547,17 @@ struct uae_prefs {
 	int cpu060_revision;
 	int fpu_model;
 	int fpu_revision;
+	int ppc_mode;
+	TCHAR ppc_model[32];
 	bool cpu_compatible;
+	bool int_no_unimplemented;
+	bool fpu_no_unimplemented;
 	bool address_space_24;
 	bool picasso96_nocustom;
 	int picasso96_modeflags;
 
+	uae_u32 z3autoconfig_start;
 	uae_u32 z3fastmem_size, z3fastmem2_size;
-	uae_u32 z3fastmem_start;
 	uae_u32 z3chipmem_size;
 	uae_u32 z3chipmem_start;
 	uae_u32 fastmem_size, fastmem2_size;
@@ -422,10 +566,18 @@ struct uae_prefs {
 	uae_u32 bogomem_size;
 	uae_u32 mbresmem_low_size;
 	uae_u32 mbresmem_high_size;
+	uae_u32 mem25bit_size;
 	uae_u32 rtgmem_size;
+	int cpuboard_type;
+	int cpuboard_subtype;
+	int cpuboard_settings;
+	uae_u32 cpuboardmem1_size;
+	uae_u32 cpuboardmem2_size;
+	int ppc_implementation;
 	bool rtg_hardwareinterrupt;
 	bool rtg_hardwaresprite;
 	int rtgmem_type;
+	bool rtg_more_compatible;
 	uae_u32 custom_memory_addrs[MAX_CUSTOM_MEMORY_ADDRS];
 	uae_u32 custom_memory_sizes[MAX_CUSTOM_MEMORY_ADDRS];
 
@@ -436,17 +588,24 @@ struct uae_prefs {
 	int uae_hide;
 	bool clipboard_sharing;
 	bool native_code;
+	bool uae_hide_autoconfig;
+	int z3_mapping_mode;
+	bool sound_toccata;
+	bool sound_toccata_mixer;
 
 	int mountitems;
-	struct uaedev_config_info mountconfig[MOUNT_CONFIG_SIZE];
+	struct uaedev_config_data mountconfig[MOUNT_CONFIG_SIZE];
 
 	int nr_floppies;
 	struct floppyslot floppyslots[4];
 	bool floppy_read_only;
 	TCHAR dfxlist[MAX_SPARE_DRIVES][MAX_DPATH];
-	int dfxclickvolume;
+	int dfxclickvolume_disk[4];
+	int dfxclickvolume_empty[4];
 	int dfxclickchannelmask;
 
+	TCHAR luafiles[MAX_LUA_STATES][MAX_DPATH];
+
 	/* Target specific options */
 
 	bool win32_middle_mouse;
@@ -466,12 +625,13 @@ struct uae_prefs {
 	int win32_inactive_priority;
 	bool win32_inactive_pause;
 	bool win32_inactive_nosound;
+	int win32_inactive_input;
 	int win32_iconified_priority;
 	bool win32_iconified_pause;
 	bool win32_iconified_nosound;
+	int win32_iconified_input;
 
 	bool win32_rtgmatchdepth;
-	bool win32_rtgscaleifsmall;
 	bool win32_rtgallowscaling;
 	int win32_rtgscaleaspectratio;
 	int win32_rtgvblankrate;
@@ -499,7 +659,12 @@ struct uae_prefs {
 	TCHAR win32_guipage[32];
 	TCHAR win32_guiactivepage[32];
 	bool win32_filesystem_mangle_reserved_names;
-
+#ifdef WITH_SLIRP
+#ifdef FSUAE
+	int slirp_implementation;
+#endif
+	struct slirp_redir slirp_redirs[MAX_SLIRP_REDIRS];
+#endif
 	int statecapturerate, statecapturebuffersize;
 
 	/* input */
@@ -515,9 +680,11 @@ struct uae_prefs {
 	int input_autofire_linecnt;
 	int input_mouse_speed;
 	int input_tablet;
+	bool tablet_library;
 	bool input_magic_mouse;
 	int input_magic_mouse_cursor;
 	int input_keyboard_type;
+	int input_autoswitch;
 	struct uae_input_device joystick_settings[MAX_INPUT_SETTINGS][MAX_INPUT_DEVICES];
 	struct uae_input_device mouse_settings[MAX_INPUT_SETTINGS][MAX_INPUT_DEVICES];
 	struct uae_input_device keyboard_settings[MAX_INPUT_SETTINGS][MAX_INPUT_DEVICES];
@@ -529,6 +696,7 @@ struct uae_prefs {
 
 extern int config_changed;
 extern void config_check_vsync (void);
+extern void set_config_changed (void);
 
 /* Contains the filename of .uaerc */
 extern TCHAR optionsfile[];
@@ -550,11 +718,13 @@ extern void cfgfile_target_write_str (struct zfile *f, const TCHAR *option, cons
 extern void cfgfile_target_dwrite_str (struct zfile *f, const TCHAR *option, const TCHAR *value);
 
 extern void cfgfile_backup (const TCHAR *path);
-extern struct uaedev_config_info *add_filesys_config (struct uae_prefs *p, int index,
-	const TCHAR *devname, const TCHAR *volname, const TCHAR *rootdir, bool readonly,
-	int cyls, int secspertrack, int surfaces, int reserved,
-	int blocksize, int bootpri, const TCHAR *filesysdir, int hdc, int flags,
-	int pcyls, int pheads, int psecs);
+extern struct uaedev_config_data *add_filesys_config (struct uae_prefs *p, int index, struct uaedev_config_info*);
+extern bool get_hd_geometry (struct uaedev_config_info *);
+extern void uci_set_defaults (struct uaedev_config_info *uci, bool rdb);
+
+extern void error_log (const TCHAR*, ...);
+extern TCHAR *get_error_log (void);
+extern bool is_error_log (void);
 
 extern void default_prefs (struct uae_prefs *, int);
 extern void discard_prefs (struct uae_prefs *, int);
@@ -581,20 +751,23 @@ extern int cfgfile_load (struct uae_prefs *p, const TCHAR *filename, int *type,
 extern int cfgfile_save (struct uae_prefs *p, const TCHAR *filename, int);
 extern void cfgfile_parse_line (struct uae_prefs *p, TCHAR *, int);
 extern void cfgfile_parse_lines (struct uae_prefs *p, const TCHAR *, int);
-extern int cfgfile_parse_option (struct uae_prefs *p, TCHAR *option, TCHAR *value, int);
+extern int cfgfile_parse_option (struct uae_prefs *p, const TCHAR *option, TCHAR *value, int);
 extern int cfgfile_get_description (const TCHAR *filename, TCHAR *description, TCHAR *hostlink, TCHAR *hardwarelink, int *type);
 extern void cfgfile_show_usage (void);
+extern int cfgfile_searchconfig(const TCHAR *in, int index, TCHAR *out, int outsize);
 extern uae_u32 cfgfile_uaelib (int mode, uae_u32 name, uae_u32 dst, uae_u32 maxlen);
 extern uae_u32 cfgfile_uaelib_modify (uae_u32 mode, uae_u32 parms, uae_u32 size, uae_u32 out, uae_u32 outsize);
-extern uae_u32 cfgfile_modify (uae_u32 index, TCHAR *parms, uae_u32 size, TCHAR *out, uae_u32 outsize);
+extern uae_u32 cfgfile_modify (uae_u32 index, const TCHAR *parms, uae_u32 size, TCHAR *out, uae_u32 outsize);
 extern void cfgfile_addcfgparam (TCHAR *);
 extern int built_in_prefs (struct uae_prefs *p, int model, int config, int compa, int romcheck);
 extern int built_in_chipset_prefs (struct uae_prefs *p);
+extern int built_in_cpuboard_prefs(struct uae_prefs *p);
 extern int cmdlineparser (const TCHAR *s, TCHAR *outp[], int max);
 extern int cfgfile_configuration_change (int);
 extern void fixup_prefs_dimensions (struct uae_prefs *prefs);
 extern void fixup_prefs (struct uae_prefs *prefs);
 extern void fixup_cpu (struct uae_prefs *prefs);
+bool cfgfile_board_enabled(struct uae_prefs *p, int romtype, int devnum);
 
 extern void check_prefs_changed_custom (void);
 extern void check_prefs_changed_cpu (void);
@@ -606,3 +779,5 @@ extern struct uae_prefs currprefs, changed_prefs;
 
 extern int machdep_init (void);
 extern void machdep_free (void);
+
+#endif // UAE_OPTIONS_H
