@@ -1,5 +1,9 @@
+#ifndef UAE_UAEIPC_H
+#define UAE_UAEIPC_H
 
-#define COMPIPENAME _T("WinUAE_COM")
+#ifdef FSUAE // NL
+#include "uae/types.h"
+#endif
 
 extern void *createIPC (const TCHAR *name, int);
 extern void closeIPC (void*);
@@ -8,3 +12,5 @@ extern void *geteventhandleIPC (void*);
 extern int sendBinIPC (void*, uae_u8 *msg, int len);
 extern int sendIPC (void*, TCHAR *msg);
 extern int isIPC (const TCHAR *pipename);
+
+#endif // UAE_UAEIPC_H
