@@ -15,7 +15,6 @@
 #include "uae.h"
 #include "gensound.h"
 #include "audio.h"
-#include "sounddep/sound.h"
 #include "events.h"
 #include "uae/memory.h"
 #include "custom.h"
@@ -29,33 +28,30 @@
 #include "gui.h"
 #include "zfile.h"
 #include "autoconf.h"
-#include "traps.h"
-#include "osemu.h"
 #include "picasso96.h"
-#include "bsdsocket.h"
-#include "uaeexe.h"
 #include "native2amiga.h"
-#include "scsidev.h"
-#include "uaeserial.h"
-#include "akiko.h"
-#include "cdtv.h"
 #include "savestate.h"
 #include "filesys.h"
-#include "parallel.h"
-#include "a2091.h"
-#include "a2065.h"
-#include "ncr_scsi.h"
-#include "scsi.h"
-#include "sana2.h"
 #include "blkdev.h"
-#include "gfxfilter.h"
-#include "uaeresource.h"
-#include "dongle.h"
-#include "sampler.h"
 #include "consolehook.h"
+#include "gfxboard.h"
+#include "luascript.h"
+#include "uaenative.h"
+#include "tabletlibrary.h"
+#include "cpuboard.h"
+#include "uae/ppc.h"
+#include "devices.h"
+#include "jit/compemu.h"
 #ifdef RETROPLATFORM
 #include "rp.h"
 #endif
+
+#ifdef FSUAE // NL
+// SDL is not used directly here by FS-UAE, but USE_SDL is still defined when
+// compiling FS-UAE so we explicitly undef it here.
+#undef USE_SDL
+#endif // NL
+
 #ifdef USE_SDL
 #include "SDL.h"
 #endif
@@ -109,7 +105,7 @@ uae_u32 uaerandgetseed (void)
 void my_trim (TCHAR *s)
 {
 	int len;
-	while (_tcscspn (s, _T("\t \r\n")) == 0)
+	while (_tcslen (s) > 0 && _tcscspn (s, _T("\t \r\n")) == 0)
 		memmove (s, s + 1, (_tcslen (s + 1) + 1) * sizeof (TCHAR));
 	len = _tcslen (s);
 	while (len > 0 && _tcscspn (s + len - 1, _T("\t \r\n")) == 0)
@@ -142,6 +138,8 @@ void discard_prefs (struct uae_prefs *p, int type)
 		xfree (s->option);
 		xfree (s);
 	}
+	p->all_lines = NULL;
+	currprefs.all_lines = changed_prefs.all_lines = NULL;
 #ifdef FILESYS
 	filesys_cleanup ();
 #endif
@@ -149,14 +147,24 @@ void discard_prefs (struct uae_prefs *p, int type)
 
 static void fixup_prefs_dim2 (struct wh *wh)
 {
-	if (wh->width < 160)
+	if (wh->special)
+		return;
+	if (wh->width < 160) {
+		error_log (_T("Width (%d) must be at least 160."), wh->width);
 		wh->width = 160;
-	if (wh->height < 128)
+	}
+	if (wh->height < 128) {
+		error_log (_T("Height (%d) must be at least 128."), wh->height);
 		wh->height = 128;
-	if (wh->width > 3072)
-		wh->width = 3072;
-	if (wh->height > 2048)
-		wh->height = 2048;
+	}
+	if (wh->width > max_uae_width) {
+		error_log (_T("Width (%d) max is %d."), wh->width, max_uae_width);
+		wh->width = max_uae_width;
+	}
+	if (wh->height > max_uae_height) {
+		error_log (_T("Height (%d) max is %d."), wh->height, max_uae_height);
+		wh->height = max_uae_height;
+	}
 }
 
 void fixup_prefs_dimensions (struct uae_prefs *prefs)
@@ -169,6 +177,7 @@ void fixup_prefs_dimensions (struct uae_prefs *prefs)
 	for (int i = 0; i < 2; i++) {
 		struct apmode *ap = &prefs->gfx_apmode[i];
 		ap->gfx_vflip = 0;
+		ap->gfx_strobo = false;
 		if (ap->gfx_vsync) {
 			if (ap->gfx_vsyncmode) {
 				// low latency vsync: no flip only if no-buffer
@@ -176,75 +185,145 @@ void fixup_prefs_dimensions (struct uae_prefs *prefs)
 					ap->gfx_vflip = 1;
 				if (!i && ap->gfx_backbuffers == 2)
 					ap->gfx_vflip = 1;
+				ap->gfx_strobo = prefs->lightboost_strobo;
 			} else {
 				// legacy vsync: always wait for flip
 				ap->gfx_vflip = -1;
 				if (prefs->gfx_api && ap->gfx_backbuffers < 1)
 					ap->gfx_backbuffers = 1;
+				if (ap->gfx_vflip)
+					ap->gfx_strobo = prefs->lightboost_strobo;;
 			}
 		} else {
 			// no vsync: wait if triple bufferirng
 			if (ap->gfx_backbuffers >= 2)
 				ap->gfx_vflip = -1;
 		}
+		if (prefs->gf[i].gfx_filter == 0 && ((prefs->gf[i].gfx_filter_autoscale && !prefs->gfx_api) || (prefs->gfx_apmode[APMODE_NATIVE].gfx_vsyncmode))) {
+			prefs->gf[i].gfx_filter = 1;
+		}
+		if (i == 0) {
+			if (prefs->gf[i].gfx_filter == 0 && prefs->monitoremu) {
+				error_log(_T("A2024 and Graffiti require at least null filter enabled."));
+				prefs->gf[i].gfx_filter = 1;
+			}
+			if (prefs->gf[i].gfx_filter == 0 && prefs->cs_cd32fmv) {
+				error_log(_T("CD32 MPEG module overlay support require at least null filter enabled."));
+				prefs->gf[i].gfx_filter = 1;
+			}
+		}
 	}
-
-	if (prefs->gfx_filter == 0 && ((prefs->gfx_filter_autoscale && !prefs->gfx_api) || (prefs->gfx_apmode[0].gfx_vsyncmode)))
-		prefs->gfx_filter = 1;
-	if (prefs->gfx_filter == 0 && prefs->monitoremu)
-		prefs->gfx_filter = 1;
 }
 
 void fixup_cpu (struct uae_prefs *p)
 {
+#ifdef FSUAE
+	write_log("fixup_cpu\n");
+#endif
 	if (p->cpu_frequency == 1000000)
 		p->cpu_frequency = 0;
+
+	if (p->cpu_model >= 68020 && p->cpuboard_type && p->address_space_24 && cpuboard_32bit(p)) {
+		error_log (_T("24-bit address space is not supported with selected accelerator board configuration."));
+		p->address_space_24 = 0;
+	}
+	if (p->cpu_model >= 68040 && p->address_space_24) {
+		error_log (_T("24-bit address space is not supported with 68040/060 configurations."));
+		p->address_space_24 = 0;
+	}
+	if (p->cpu_model < 68020 && p->fpu_model && (p->cpu_compatible || p->cpu_cycle_exact)) {
+		error_log (_T("FPU is not supported with 68000/010 configurations."));
+		p->fpu_model = 0;
+	}
+
 	switch (p->cpu_model)
 	{
 	case 68000:
-		p->address_space_24 = 1;
-		if (p->cpu_compatible || p->cpu_cycle_exact)
-			p->fpu_model = 0;
 		break;
 	case 68010:
-		p->address_space_24 = 1;
-		if (p->cpu_compatible || p->cpu_cycle_exact)
-			p->fpu_model = 0;
 		break;
 	case 68020:
 		break;
 	case 68030:
-		p->address_space_24 = 0;
 		break;
 	case 68040:
-		p->address_space_24 = 0;
 		if (p->fpu_model)
 			p->fpu_model = 68040;
 		break;
 	case 68060:
-		p->address_space_24 = 0;
 		if (p->fpu_model)
 			p->fpu_model = 68060;
 		break;
 	}
 
-	if (p->cpu_model != 68040)
-		p->mmu_model = 0;
+	// 1 = "automatic" PPC config
+	if (p->ppc_mode == 1) {
+		cpuboard_setboard(p,  BOARD_CYBERSTORM, BOARD_CYBERSTORM_SUB_PPC);
+		if (p->cs_compatible == CP_A1200) {
+			cpuboard_setboard(p,  BOARD_BLIZZARD, BOARD_BLIZZARD_SUB_PPC);
+		} else if (p->cs_compatible != CP_A4000 && p->cs_compatible != CP_A4000T && p->cs_compatible != CP_A3000 && p->cs_compatible != CP_A3000T) {
+			if ((p->cs_ide == IDE_A600A1200 || p->cs_pcmcia) && p->cs_mbdmac <= 0) {
+				cpuboard_setboard(p,  BOARD_BLIZZARD, BOARD_BLIZZARD_SUB_PPC);
+			}
+		}
+		if (p->cpuboardmem1_size < 8 * 1024 * 1024)
+			p->cpuboardmem1_size = 8 * 1024 * 1024;
+	}
 
-	if (p->cachesize && p->cpu_cycle_exact)
+	if (p->cpu_model < 68020 && p->cachesize) {
 		p->cachesize = 0;
+		error_log (_T("JIT requires 68020 or better CPU."));
+	}
 
-	if (p->cpu_cycle_exact && p->m68k_speed < 0)
+	if (p->cpu_model >= 68040 && p->cachesize && p->cpu_compatible)
+		p->cpu_compatible = false;
+
+	if ((p->cpu_model < 68030 || p->cachesize) && p->mmu_model) {
+		error_log (_T("MMU emulation requires 68030/040/060 and it is not JIT compatible."));
+		p->mmu_model = 0;
+	}
+
+	if (p->cachesize && p->cpu_cycle_exact) {
+		error_log (_T("JIT and cycle-exact can't be enabled simultaneously."));
+		p->cachesize = 0;
+	}
+	if (p->cachesize && (p->fpu_no_unimplemented || p->int_no_unimplemented)) {
+		error_log (_T("JIT is not compatible with unimplemented CPU/FPU instruction emulation."));
+		p->fpu_no_unimplemented = p->int_no_unimplemented = false;
+	}
+
+#if 0
+	if (p->cpu_cycle_exact && p->m68k_speed < 0 && currprefs.cpu_model <= 68020)
 		p->m68k_speed = 0;
+#endif
 
-	if (p->immediate_blits && p->blitter_cycle_exact)
+#if 0
+	if (p->immediate_blits && p->blitter_cycle_exact) {
+		error_log (_T("Cycle-exact and immediate blitter can't be enabled simultaneously.\n"));
 		p->immediate_blits = false;
-	if (p->immediate_blits && p->waiting_blits)
+	}
+#endif
+	if (p->immediate_blits && p->waiting_blits) {
+		error_log (_T("Immediate blitter and waiting blits can't be enabled simultaneously.\n"));
 		p->waiting_blits = 0;
+	}
+	if (p->cpu_cycle_exact)
+		p->cpu_compatible = true;
+
+	if (p->cpu_cycle_exact && p->produce_sound == 0) {
+		p->produce_sound = 1;
+		error_log(_T("Cycle-exact mode requires at least Disabled but emulated sound setting."));
+	}
+
+	if (p->cpuboard_type && cpuboard_jitdirectompatible(p) && !p->comptrustbyte) {
+		error_log(_T("JIT direct is not compatible with emulated Blizzard accelerator boards."));
+		p->comptrustbyte = 1;
+		p->comptrustlong = 1;
+		p->comptrustlong = 1;
+		p->comptrustnaddr = 1;
+	}
 }
 
-
-
 void fixup_prefs (struct uae_prefs *p)
 {
 	int err = 0;
@@ -252,99 +331,176 @@ void fixup_prefs (struct uae_prefs *p)
 	built_in_chipset_prefs (p);
 	fixup_cpu (p);
 
+
+	if (p->cpuboard_type && p->cpuboardmem1_size > cpuboard_maxmemory(p)) {
+		error_log(_T("Unsupported accelerator board memory size %d (0x%x).\n"), p->cpuboardmem1_size, p->cpuboardmem1_size);
+		p->cpuboardmem1_size = cpuboard_maxmemory(p);
+	}
+	if (cpuboard_memorytype(p) == BOARD_MEMORY_HIGHMEM) {
+		p->mbresmem_high_size = p->cpuboardmem1_size;
+	} else if (cpuboard_memorytype(p) == BOARD_MEMORY_Z2) {
+		p->fastmem2_size = p->cpuboardmem1_size;
+	} else if (cpuboard_memorytype(p) == BOARD_MEMORY_25BITMEM) {
+		p->mem25bit_size = p->cpuboardmem1_size;
+	}
+
 	if (((p->chipmem_size & (p->chipmem_size - 1)) != 0 && p->chipmem_size != 0x180000)
 		|| p->chipmem_size < 0x20000
 		|| p->chipmem_size > 0x800000)
 	{
-		write_log (_T("Unsupported chipmem size %x!\n"), p->chipmem_size);
+		error_log (_T("Unsupported chipmem size %d (0x%x)."), p->chipmem_size, p->chipmem_size);
 		p->chipmem_size = 0x200000;
 		err = 1;
 	}
+
 	if ((p->fastmem_size & (p->fastmem_size - 1)) != 0
-		|| (p->fastmem_size != 0 && (p->fastmem_size < 0x100000 || p->fastmem_size > 0x800000)))
+		|| (p->fastmem_size != 0 && (p->fastmem_size < 0x10000 || p->fastmem_size > 0x800000)))
 	{
-		write_log (_T("Unsupported fastmem size %x!\n"), p->fastmem_size);
+		error_log (_T("Unsupported fastmem size %d (0x%x)."), p->fastmem_size, p->fastmem_size);
+		p->fastmem_size = 0;
 		err = 1;
 	}
-	if ((p->rtgmem_size & (p->rtgmem_size - 1)) != 0
-		|| (p->rtgmem_size != 0 && (p->rtgmem_size < 0x100000 || p->rtgmem_size > max_z3fastmem)))
+	if ((p->fastmem2_size & (p->fastmem2_size - 1)) != 0 || (p->fastmem2_size != 0 && (p->fastmem2_size < 0x10000 || p->fastmem2_size > 0x800000)))
 	{
-		write_log (_T("Unsupported graphics card memory size %x (%x)!\n"), p->rtgmem_size, max_z3fastmem);
+		error_log (_T("Unsupported fastmem2 size %d (0x%x)."), p->fastmem2_size, p->fastmem2_size);
+		p->fastmem2_size = 0;
+		err = 1;
+	}
+	if (p->cachesize) {
+		if (p->fastmem_size + p->fastmem2_size > 0x800000) {
+			error_log (_T("Unsupported fastmem2 size %d (0x%x)."), p->fastmem2_size, p->fastmem2_size);
+			err = 1;
+		}
+		if (p->fastmem2_size > p->fastmem_size && p->fastmem_size > 0) {
+			error_log (_T("Fastmem2 size can't be larger than fastmem1 if JIT is enabled."));
+			p->fastmem2_size = 0;
+			err = 1;
+		}
+	}
+
+	if (p->rtgmem_size > max_z3fastmem && p->rtgmem_type == GFXBOARD_UAE_Z3) {
+		error_log (_T("Graphics card memory size %d (0x%x) larger than maximum reserved %d (0x%x)."), p->rtgmem_size, p->rtgmem_size, max_z3fastmem, max_z3fastmem);
+		p->rtgmem_size = max_z3fastmem;
+		err = 1;
+	}
+
+	if ((p->rtgmem_size & (p->rtgmem_size - 1)) != 0 || (p->rtgmem_size != 0 && (p->rtgmem_size < 0x100000))) {
+		error_log (_T("Unsupported graphics card memory size %d (0x%x)."), p->rtgmem_size, p->rtgmem_size);
 		if (p->rtgmem_size > max_z3fastmem)
 			p->rtgmem_size = max_z3fastmem;
 		else
 			p->rtgmem_size = 0;
 		err = 1;
 	}
+	
+	if (p->z3fastmem_size > max_z3fastmem) {
+		error_log (_T("Zorro III fastmem size %d (0x%x) larger than max reserved %d (0x%x)."), p->z3fastmem_size, p->z3fastmem_size, max_z3fastmem, max_z3fastmem);
+		p->z3fastmem_size = max_z3fastmem;
+		err = 1;
+	}
+	if ((p->z3fastmem_size & (p->z3fastmem_size - 1)) != 0 || (p->z3fastmem_size != 0 && p->z3fastmem_size < 0x100000))
+	{
+		error_log (_T("Unsupported Zorro III fastmem size %d (0x%x)."), p->z3fastmem_size, p->z3fastmem_size);
+		p->z3fastmem_size = 0;
+		err = 1;
+	}
 
-	if ((p->z3fastmem_size & (p->z3fastmem_size - 1)) != 0
-		|| (p->z3fastmem_size != 0 && (p->z3fastmem_size < 0x100000 || p->z3fastmem_size > max_z3fastmem)))
-	{
-		write_log (_T("Unsupported Zorro III fastmem size %x (%x)!\n"), p->z3fastmem_size, max_z3fastmem);
-		if (p->z3fastmem_size > max_z3fastmem)
-			p->z3fastmem_size = max_z3fastmem;
-		else
-			p->z3fastmem_size = 0;
+	if (p->z3fastmem2_size > max_z3fastmem) {
+		error_log (_T("Zorro III fastmem2 size %d (0x%x) larger than max reserved %d (0x%x)."), p->z3fastmem2_size, p->z3fastmem2_size, max_z3fastmem, max_z3fastmem);
+		p->z3fastmem2_size = max_z3fastmem;
 		err = 1;
 	}
-	if ((p->z3fastmem2_size & (p->z3fastmem2_size - 1)) != 0
-		|| (p->z3fastmem2_size != 0 && (p->z3fastmem2_size < 0x100000 || p->z3fastmem2_size > max_z3fastmem)))
+	if ((p->z3fastmem2_size & (p->z3fastmem2_size - 1)) != 0 || (p->z3fastmem2_size != 0 && p->z3fastmem2_size < 0x100000))
 	{
-		write_log (_T("Unsupported Zorro III fastmem size %x (%x)!\n"), p->z3fastmem2_size, max_z3fastmem);
-		if (p->z3fastmem2_size > max_z3fastmem)
-			p->z3fastmem2_size = max_z3fastmem;
-		else
-			p->z3fastmem2_size = 0;
+		error_log (_T("Unsupported Zorro III fastmem2 size %x (%x)."), p->z3fastmem2_size, p->z3fastmem2_size);
+		p->z3fastmem2_size = 0;
 		err = 1;
 	}
-	p->z3fastmem_start &= ~0xffff;
-	if (p->z3fastmem_start < 0x1000000)
-		p->z3fastmem_start = 0x1000000;
-	if ((p->z3chipmem_size & (p->z3chipmem_size - 1)) != 0
-		|| (p->z3chipmem_size != 0 && (p->z3chipmem_size < 0x100000 || p->z3chipmem_size > max_z3fastmem)))
+
+	p->z3autoconfig_start &= ~0xffff;
+	if (p->z3autoconfig_start < 0x1000000)
+		p->z3autoconfig_start = 0x1000000;
+
+	if (p->z3chipmem_size > max_z3fastmem) {
+		error_log (_T("Zorro III fake chipmem size %d (0x%x) larger than max reserved %d (0x%x)."), p->z3chipmem_size, p->z3chipmem_size, max_z3fastmem, max_z3fastmem);
+		p->z3chipmem_size = max_z3fastmem;
+		err = 1;
+	}
+	if (((p->z3chipmem_size & (p->z3chipmem_size - 1)) != 0 &&  p->z3chipmem_size != 0x18000000 && p->z3chipmem_size != 0x30000000) || (p->z3chipmem_size != 0 && p->z3chipmem_size < 0x100000))
 	{
-		write_log (_T("Unsupported Zorro III fake chipmem size %x (%x)!\n"), p->z3chipmem_size, max_z3fastmem);
-		if (p->z3chipmem_size > max_z3fastmem)
-			p->z3chipmem_size = max_z3fastmem;
-		else
-			p->z3chipmem_size = 0;
+		error_log (_T("Unsupported 32-bit chipmem size %d (0x%x)."), p->z3chipmem_size, p->z3chipmem_size);
+		p->z3chipmem_size = 0;
 		err = 1;
 	}
 
 	if (p->address_space_24 && (p->z3fastmem_size != 0 || p->z3fastmem2_size != 0 || p->z3chipmem_size != 0)) {
 		p->z3fastmem_size = p->z3fastmem2_size = p->z3chipmem_size = 0;
-		write_log (_T("Can't use a graphics card or 32-bit memory when using a 24 bit\naddress space.\n"));
+		error_log (_T("Can't use a Z3 graphics card or 32-bit memory when using a 24 bit address space."));
 	}
+
 	if (p->bogomem_size != 0 && p->bogomem_size != 0x80000 && p->bogomem_size != 0x100000 && p->bogomem_size != 0x180000 && p->bogomem_size != 0x1c0000) {
+		error_log (_T("Unsupported bogomem size %d (0x%x)"), p->bogomem_size, p->bogomem_size);
 		p->bogomem_size = 0;
-		write_log (_T("Unsupported bogomem size!\n"));
 		err = 1;
 	}
+
 	if (p->bogomem_size > 0x180000 && (p->cs_fatgaryrev >= 0 || p->cs_ide || p->cs_ramseyrev >= 0)) {
 		p->bogomem_size = 0x180000;
-		write_log (_T("Possible Gayle bogomem conflict fixed\n"));
+		error_log (_T("Possible Gayle bogomem conflict fixed."));
 	}
-	if (p->chipmem_size > 0x200000 && p->fastmem_size != 0) {
-		write_log (_T("You can't use fastmem and more than 2MB chip at the same time!\n"));
-		p->fastmem_size = 0;
+	if (p->chipmem_size > 0x200000 && p->fastmem_size > 262144) {
+		error_log(_T("You can't use fastmem and more than 2MB chip at the same time."));
+		p->chipmem_size = 0x200000;
 		err = 1;
 	}
+	if (p->chipmem_size > 0x200000 && p->rtgmem_size && !gfxboard_is_z3(p->rtgmem_type)) {
+		error_log(_T("You can't use Zorro II RTG and more than 2MB chip at the same time."));
+		p->chipmem_size = 0x200000;
+		err = 1;
+	}
+	if (p->mem25bit_size > 128 * 1024 * 1024 || (p->mem25bit_size & 0xfffff)) {
+		p->mem25bit_size = 0;
+		error_log (_T("Unsupported 25bit RAM size"));
+	}
 	if (p->mbresmem_low_size > 0x04000000 || (p->mbresmem_low_size & 0xfffff)) {
 		p->mbresmem_low_size = 0;
-		write_log (_T("Unsupported A3000 MB RAM size\n"));
+		error_log (_T("Unsupported Mainboard RAM size"));
 	}
-	if (p->mbresmem_high_size > 0x04000000 || (p->mbresmem_high_size & 0xfffff)) {
+	if (p->mbresmem_high_size > 0x08000000 || (p->mbresmem_high_size & 0xfffff)) {
 		p->mbresmem_high_size = 0;
-		write_log (_T("Unsupported Motherboard RAM size\n"));
+		error_log (_T("Unsupported CPU Board RAM size."));
 	}
 
-	if (p->address_space_24 && p->rtgmem_size)
-		p->rtgmem_type = 0;
-	if (!p->rtgmem_type && (p->chipmem_size > 2 * 1024 * 1024 || getz2size (p) > 8 * 1024 * 1024 || getz2size (p) < 0)) {
+	if (p->rtgmem_type >= GFXBOARD_HARDWARE) {
+		if (gfxboard_get_vram_min(p->rtgmem_type) > 0 && p->rtgmem_size < gfxboard_get_vram_min (p->rtgmem_type)) {
+			error_log(_T("Graphics card memory size %d (0x%x) smaller than minimum hardware supported %d (0x%x)."),
+				p->rtgmem_size, p->rtgmem_size, gfxboard_get_vram_min(p->rtgmem_type), gfxboard_get_vram_min(p->rtgmem_type));
+			p->rtgmem_size = gfxboard_get_vram_min (p->rtgmem_type);
+		}
+		if (p->address_space_24 && gfxboard_is_z3 (p->rtgmem_type)) {
+			p->rtgmem_type = GFXBOARD_UAE_Z2;
+			p->rtgmem_size = 0;
+			error_log (_T("Z3 RTG and 24-bit address space are not compatible."));
+		}
+		if (gfxboard_get_vram_max(p->rtgmem_type) > 0 && p->rtgmem_size > gfxboard_get_vram_max(p->rtgmem_type)) {
+			error_log(_T("Graphics card memory size %d (0x%x) larger than maximum hardware supported %d (0x%x)."),
+				p->rtgmem_size, p->rtgmem_size, gfxboard_get_vram_max(p->rtgmem_type), gfxboard_get_vram_max(p->rtgmem_type));
+			p->rtgmem_size = gfxboard_get_vram_max(p->rtgmem_type);
+		}
+	}
+	if (p->address_space_24 && p->rtgmem_size && p->rtgmem_type == GFXBOARD_UAE_Z3) {
+		error_log (_T("Z3 RTG and 24bit address space are not compatible."));
+		p->rtgmem_type = GFXBOARD_UAE_Z2;
+	}
+	if (p->rtgmem_type == GFXBOARD_UAE_Z2 && (p->chipmem_size > 2 * 1024 * 1024 || getz2size (p) > 8 * 1024 * 1024 || getz2size (p) < 0)) {
 		p->rtgmem_size = 0;
-		write_log (_T("Too large Z2 RTG memory size\n"));
+		error_log (_T("Too large Z2 RTG memory size."));
 	}
 
+	if (p->cs_z3autoconfig && p->address_space_24) {
+		p->cs_z3autoconfig = false;
+		error_log (_T("Z3 autoconfig and 24bit address space are not compatible."));
+	}
 
 #if 0
 	if (p->m68k_speed < -1 || p->m68k_speed > 20) {
@@ -355,47 +511,48 @@ void fixup_prefs (struct uae_prefs *p)
 #endif
 
 	if (p->produce_sound < 0 || p->produce_sound > 3) {
-		write_log (_T("Bad value for -S parameter: enable value must be within 0..3\n"));
+		error_log (_T("Bad value for -S parameter: enable value must be within 0..3."));
 		p->produce_sound = 0;
 		err = 1;
 	}
 	if (p->comptrustbyte < 0 || p->comptrustbyte > 3) {
-		write_log (_T("Bad value for comptrustbyte parameter: value must be within 0..2\n"));
+		error_log (_T("Bad value for comptrustbyte parameter: value must be within 0..2."));
 		p->comptrustbyte = 1;
 		err = 1;
 	}
 	if (p->comptrustword < 0 || p->comptrustword > 3) {
-		write_log (_T("Bad value for comptrustword parameter: value must be within 0..2\n"));
+		error_log (_T("Bad value for comptrustword parameter: value must be within 0..2."));
 		p->comptrustword = 1;
 		err = 1;
 	}
 	if (p->comptrustlong < 0 || p->comptrustlong > 3) {
-		write_log (_T("Bad value for comptrustlong parameter: value must be within 0..2\n"));
+		error_log (_T("Bad value for comptrustlong parameter: value must be within 0..2."));
 		p->comptrustlong = 1;
 		err = 1;
 	}
 	if (p->comptrustnaddr < 0 || p->comptrustnaddr > 3) {
-		write_log (_T("Bad value for comptrustnaddr parameter: value must be within 0..2\n"));
+		error_log (_T("Bad value for comptrustnaddr parameter: value must be within 0..2."));
 		p->comptrustnaddr = 1;
 		err = 1;
 	}
 	if (p->cachesize < 0 || p->cachesize > 16384) {
-		write_log (_T("Bad value for cachesize parameter: value must be within 0..16384\n"));
+		error_log (_T("Bad value for cachesize parameter: value must be within 0..16384."));
 		p->cachesize = 0;
 		err = 1;
 	}
-	if (p->z3fastmem_size > 0 && (p->address_space_24 || p->cpu_model < 68020)) {
-		write_log (_T("Z3 fast memory can't be used with a 68000/68010 emulation. It\n")
-			_T("requires a 68020 emulation. Turning off Z3 fast memory.\n"));
+	if ((p->z3fastmem_size || p->z3fastmem2_size || p->z3chipmem_size) && p->address_space_24) {
+		error_log (_T("Z3 fast memory can't be used if address space is 24-bit."));
 		p->z3fastmem_size = 0;
+		p->z3fastmem2_size = 0;
+		p->z3chipmem_size = 0;
 		err = 1;
 	}
-	if (p->rtgmem_size > 0 && p->rtgmem_type && (p->cpu_model < 68020 || p->address_space_24)) {
-		write_log (_T("RTG can't be used with a 68000/68010 or 68EC020 emulation. It\n")
-			_T("requires a 68020 emulation. Turning off RTG.\n"));
+	if ((p->rtgmem_size > 0 && p->rtgmem_type == GFXBOARD_UAE_Z3) && p->address_space_24) {
+		error_log (_T("UAEGFX RTG can't be used if address space is 24-bit."));
 		p->rtgmem_size = 0;
 		err = 1;
 	}
+
 #if !defined (BSDSOCKET)
 	if (p->socket_emu) {
 		write_log (_T("Compile-time option of BSDSOCKET_SUPPORTED was not enabled.  You can't use bsd-socket emulation.\n"));
@@ -405,22 +562,24 @@ void fixup_prefs (struct uae_prefs *p)
 #endif
 
 	if (p->nr_floppies < 0 || p->nr_floppies > 4) {
-		write_log (_T("Invalid number of floppies.  Using 4.\n"));
-		p->nr_floppies = 4;
+		error_log (_T("Invalid number of floppies.  Using 2."));
+		p->nr_floppies = 2;
 		p->floppyslots[0].dfxtype = 0;
 		p->floppyslots[1].dfxtype = 0;
-		p->floppyslots[2].dfxtype = 0;
-		p->floppyslots[3].dfxtype = 0;
+		p->floppyslots[2].dfxtype = -1;
+		p->floppyslots[3].dfxtype = -1;
 		err = 1;
 	}
 	if (p->floppy_speed > 0 && p->floppy_speed < 10) {
+		error_log (_T("Invalid floppy speed."));
 		p->floppy_speed = 100;
 	}
 	if (p->input_mouse_speed < 1 || p->input_mouse_speed > 1000) {
+		error_log (_T("Invalid mouse speed."));
 		p->input_mouse_speed = 100;
 	}
 	if (p->collision_level < 0 || p->collision_level > 3) {
-		write_log (_T("Invalid collision support level.  Using 1.\n"));
+		error_log (_T("Invalid collision support level.  Using 1."));
 		p->collision_level = 1;
 		err = 1;
 	}
@@ -443,12 +602,19 @@ void fixup_prefs (struct uae_prefs *p)
 				p->cs_ramseyrev = 0x0f;
 		}
 	}
+	if (p->chipmem_size >= 0x100000)
+		p->cs_1mchipjumper = true;
+
 	/* Can't fit genlock and A2024 or Graffiti at the same time,
 	 * also Graffiti uses genlock audio bit as an enable signal
 	 */
-	if (p->genlock && p->monitoremu)
+	if (p->genlock && p->monitoremu) {
+		error_log (_T("Genlock and A2024 or Graffiti can't be active simultaneously."));
 		p->genlock = false;
-
+	}
+	if (p->cs_hacks) {
+		error_log (_T("chipset_hacks is nonzero (0x%04x)."), p->cs_hacks);
+	}
 
 	fixup_prefs_dimensions (p);
 
@@ -463,11 +629,11 @@ void fixup_prefs (struct uae_prefs *p)
 	p->cpu_compatible = 1;
 	p->address_space_24 = 1;
 #endif
-#if !defined (CPUEMU_11) && !defined (CPUEMU_12)
+#if !defined (CPUEMU_11) && !defined (CPUEMU_13)
 	p->cpu_compatible = 0;
 	p->address_space_24 = 0;
 #endif
-#if !defined (CPUEMU_12)
+#if !defined (CPUEMU_13)
 	p->cpu_cycle_exact = p->blitter_cycle_exact = 0;
 #endif
 #ifndef AGA
@@ -493,17 +659,36 @@ void fixup_prefs (struct uae_prefs *p)
 #if !defined (UAESERIAL)
 	p->uaeserial = 0;
 #endif
-#if defined (CPUEMU_12)
+#if defined (CPUEMU_13)
 	if (p->cpu_cycle_exact) {
-		p->gfx_framerate = 1;
-		p->cachesize = 0;
-		p->m68k_speed = 0;
+		if (p->gfx_framerate > 1) {
+			error_log (_T("Cycle-exact requires disabled frameskip."));
+			p->gfx_framerate = 1;
+		}
+		if (p->cachesize) {
+			error_log (_T("Cycle-exact and JIT can't be active simultaneously."));
+			p->cachesize = 0;
+		}
+#if 0
+		if (p->m68k_speed) {
+			error_log (_T("Adjustable CPU speed is not available in cycle-exact mode."));
+			p->m68k_speed = 0;
+		}
+#endif
 	}
 #endif
-	if (p->maprom && !p->address_space_24)
+	if (p->maprom && !p->address_space_24) {
+#ifdef FSUAE
+		write_log("MAPROM: Setting address 0x0f000000 (was 0x%08x)\n", p->maprom);
+#endif
 		p->maprom = 0x0f000000;
-	if ((p->maprom & 0xff000000) && p->address_space_24)
+	}
+	if (((p->maprom & 0xff000000) && p->address_space_24) || (p->maprom && p->mbresmem_high_size >= 0x08000000)) {
+#ifdef FSUAE
+		write_log("MAPROM: Setting address 0x00e00000 (was 0x%08x)\n", p->maprom);
+#endif
 		p->maprom = 0x00e00000;
+	}
 	if (p->tod_hack && p->cs_ciaatod == 0)
 		p->cs_ciaatod = p->ntscmode ? 2 : 1;
 
@@ -537,6 +722,9 @@ void uae_reset (int hardreset, int keyboardreset)
 
 void uae_quit (void)
 {
+#ifdef FSUAE
+	write_log("uae_quit\n");
+#endif
 	deactivate_debugger ();
 	if (quit_program != -UAE_QUIT)
 		quit_program = -UAE_QUIT;
@@ -546,6 +734,9 @@ void uae_quit (void)
 /* 0 = normal, 1 = nogui, -1 = disable nogui */
 void uae_restart (int opengui, const TCHAR *cfgfile)
 {
+#ifdef FSUAE
+	write_log("uae_restart\n");
+#endif
 	uae_quit ();
 	restart_program = opengui > 0 ? 1 : (opengui == 0 ? 2 : 3);
 	restart_config[0] = 0;
@@ -592,7 +783,7 @@ static int diskswapper_cb (struct zfile *f, void *vrsd)
 static void parse_diskswapper (const TCHAR *s)
 {
 	TCHAR *tmp = my_strdup (s);
-	TCHAR *delim = _T(",");
+	const TCHAR *delim = _T(",");
 	TCHAR *p1, *p2;
 	int num = 0;
 
@@ -641,6 +832,7 @@ static TCHAR *parsetextpath (const TCHAR *s)
 static void parse_cmdline (int argc, TCHAR **argv)
 {
 	int i;
+	bool firstconfig = true;
 
 	for (i = 1; i < argc; i++) {
 		if (!_tcsncmp (argv[i], _T("-diskswapper="), 13)) {
@@ -655,8 +847,9 @@ static void parse_cmdline (int argc, TCHAR **argv)
 		} else if (_tcsncmp (argv[i], _T("-config="), 8) == 0) {
 			TCHAR *txt = parsetextpath (argv[i] + 8);
 			currprefs.mountitems = 0;
-			target_cfgfile_load (&currprefs, txt, -1, 0);
+			target_cfgfile_load (&currprefs, txt, firstconfig ? CONFIG_TYPE_ALL : CONFIG_TYPE_HARDWARE | CONFIG_TYPE_HOST | CONFIG_TYPE_NORESET, 0);
 			xfree (txt);
+			firstconfig = false;
 		} else if (_tcsncmp (argv[i], _T("-statefile="), 11) == 0) {
 			TCHAR *txt = parsetextpath (argv[i] + 11);
 			savestate_state = STATE_DORESTORE;
@@ -669,8 +862,9 @@ static void parse_cmdline (int argc, TCHAR **argv)
 			} else {
 				TCHAR *txt = parsetextpath (argv[++i]);
 				currprefs.mountitems = 0;
-				target_cfgfile_load (&currprefs, txt, -1, 0);
+				target_cfgfile_load (&currprefs, txt, firstconfig ? CONFIG_TYPE_ALL : CONFIG_TYPE_HARDWARE | CONFIG_TYPE_HOST | CONFIG_TYPE_NORESET, 0);
 				xfree (txt);
+				firstconfig = false;
 			}
 		} else if (_tcscmp (argv[i], _T("-s")) == 0) {
 			if (i + 1 == argc)
@@ -723,12 +917,12 @@ static void parse_cmdline_and_init_file (int argc, TCHAR **argv)
 
 	_tcscat (optionsfile, restart_config);
 
-	if (! target_cfgfile_load (&currprefs, optionsfile, 0, default_config)) {
+	if (! target_cfgfile_load (&currprefs, optionsfile, CONFIG_TYPE_DEFAULT, default_config)) {
 		write_log (_T("failed to load config '%s'\n"), optionsfile);
 #ifdef OPTIONS_IN_HOME
 		/* sam: if not found in $HOME then look in current directory */
 		_tcscpy (optionsfile, restart_config);
-		target_cfgfile_load (&currprefs, optionsfile, 0, default_config);
+		target_cfgfile_load (&currprefs, optionsfile, CONFIG_TYPE_DEFAULT, default_config);
 #endif
 	}
 	fixup_prefs (&currprefs);
@@ -736,50 +930,6 @@ static void parse_cmdline_and_init_file (int argc, TCHAR **argv)
 	parse_cmdline (argc, argv);
 }
 
-void reset_all_systems (void)
-{
-	init_eventtab ();
-
-#ifdef PICASSO96
-	picasso_reset ();
-#endif
-#ifdef SCSIEMU
-	scsi_reset ();
-	scsidev_reset ();
-	scsidev_start_threads ();
-#endif
-#ifdef A2065
-	a2065_reset ();
-#endif
-#ifdef SANA2
-	netdev_reset ();
-	netdev_start_threads ();
-#endif
-#ifdef FILESYS
-	filesys_prepare_reset ();
-	filesys_reset ();
-#endif
-	init_shm ();
-	memory_reset ();
-#if defined (BSDSOCKET)
-	bsdlib_reset ();
-#endif
-#ifdef FILESYS
-	filesys_start_threads ();
-	hardfile_reset ();
-#endif
-#ifdef UAESERIAL
-	uaeserialdev_reset ();
-	uaeserialdev_start_threads ();
-#endif
-#if defined (PARALLEL_PORT)
-	initparallel ();
-#endif
-	native2amiga_reset ();
-	dongle_reset ();
-	sampler_init ();
-}
-
 /* Okay, this stuff looks strange, but it is here to encourage people who
 * port UAE to re-use as much of this code as possible. Functions that you
 * should be using are do_start_program () and do_leave_program (), as well
@@ -803,24 +953,22 @@ void do_start_program (void)
 {
 	if (quit_program == -UAE_QUIT)
 		return;
-	if (!canbang && candirect < 0)
-		candirect = 0;
-	if (canbang && candirect < 0)
-		candirect = 1;
 	/* Do a reset on startup. Whether this is elegant is debatable. */
 	inputdevice_updateconfig (&changed_prefs, &currprefs);
 	if (quit_program >= 0)
 		quit_program = UAE_RESET;
-#if (defined (_WIN32) || defined (_WIN64)) && !defined (NO_WIN32_EXCEPTION_HANDLER)
-	extern int EvalException (LPEXCEPTION_POINTERS blah, int n_except);
+#ifdef WITH_LUA
+	uae_lua_loadall ();
+#endif
+#ifdef USE_STRUCTURED_EXCEPTION_HANDLING
 	__try
 #endif
 	{
 		m68k_go (1);
 	}
-#if (defined (_WIN32) || defined (_WIN64)) && !defined (NO_WIN32_EXCEPTION_HANDLER)
+#ifdef USE_STRUCTURED_EXCEPTION_HANDLING
 #ifdef JIT
-	__except (EvalException (GetExceptionInformation (), GetExceptionCode ()))
+	__except (EvalException(GetExceptionInformation()))
 #else
 	__except (DummyException (GetExceptionInformation (), GetExceptionCode ()))
 #endif
@@ -830,48 +978,6 @@ void do_start_program (void)
 #endif
 }
 
-void do_leave_program (void)
-{
-	sampler_free ();
-	graphics_leave ();
-	inputdevice_close ();
-	DISK_free ();
-	close_sound ();
-	dump_counts ();
-#ifdef SERIAL_PORT
-	serial_exit ();
-#endif
-#ifdef CDTV
-	cdtv_free ();
-#endif
-#ifdef A2091
-	a2091_free ();
-#endif
-#ifdef NCR
-	ncr_free ();
-#endif
-#ifdef CD32
-	akiko_free ();
-#endif
-	if (! no_gui)
-		gui_exit ();
-#ifdef AUTOCONFIG
-	expansion_cleanup ();
-#endif
-#ifdef FILESYS
-	filesys_cleanup ();
-#endif
-#ifdef BSDSOCKET
-	bsdlib_reset ();
-#endif
-	device_func_reset ();
-	savestate_free ();
-	memory_cleanup ();
-	free_shm ();
-	cfgfile_addcfgparam (0);
-	machdep_free ();
-}
-
 void start_program (void)
 {
 	do_start_program ();
@@ -882,47 +988,13 @@ void leave_program (void)
 	do_leave_program ();
 }
 
-
-void virtualdevice_init (void)
-{
-#ifdef AUTOCONFIG
-	rtarea_setup ();
-#endif
-#ifdef FILESYS
-	rtarea_init ();
-	uaeres_install ();
-	hardfile_install ();
-#endif
-#ifdef SCSIEMU
-	scsi_reset ();
-	scsidev_install ();
-#endif
-#ifdef SANA2
-	netdev_install ();
-#endif
-#ifdef UAESERIAL
-	uaeserialdev_install ();
-#endif
-#ifdef AUTOCONFIG
-	expansion_init ();
-	emulib_install ();
-	uaeexe_install ();
-#endif
-#ifdef FILESYS
-	filesys_install ();
-#endif
-#if defined (BSDSOCKET)
-	bsdlib_install ();
-#endif
-}
-
 static int real_main2 (int argc, TCHAR **argv)
 {
 
 #ifdef USE_SDL
 	SDL_Init (SDL_INIT_TIMER | SDL_INIT_AUDIO | SDL_INIT_JOYSTICK | SDL_INIT_NOPARACHUTE);
 #endif
-	config_changed = 1;
+	set_config_changed ();
 	if (restart_config[0]) {
 		default_prefs (&currprefs, 0);
 		fixup_prefs (&currprefs);
@@ -971,7 +1043,7 @@ static int real_main2 (int argc, TCHAR **argv)
 	if (! no_gui) {
 		int err = gui_init ();
 		currprefs = changed_prefs;
-		config_changed = 1;
+		set_config_changed ();
 		if (err == -1) {
 			write_log (_T("Failed to initialize the GUI\n"));
 			return -1;
@@ -983,13 +1055,16 @@ static int real_main2 (int argc, TCHAR **argv)
 	memset (&gui_data, 0, sizeof gui_data);
 	gui_data.cd = -1;
 	gui_data.hd = -1;
+	gui_data.md = (currprefs.cs_cd32nvram || currprefs.cs_cdtvram) ? 0 : -1;
 	logging_init (); /* Yes, we call this twice - the first case handles when the user has loaded
 						 a config using the cmd-line.  This case handles loads through the GUI. */
 
 #ifdef NATMEM_OFFSET
 	init_shm ();
 #endif
-
+#ifdef WITH_LUA
+	uae_lua_init ();
+#endif
 #ifdef PICASSO96
 	picasso_reset ();
 #endif
@@ -1019,12 +1094,14 @@ static int real_main2 (int argc, TCHAR **argv)
 #ifdef AUTOCONFIG
 	native2amiga_install ();
 #endif
-
 	custom_init (); /* Must come after memory_init */
 #ifdef SERIAL_PORT
 	serial_init ();
 #endif
 	DISK_init ();
+#ifdef WITH_PPC
+	uae_ppc_reset(true);
+#endif
 
 	reset_frame_rate_hack ();
 	init_m68k (); /* must come after reset_frame_rate_hack (); */
