@@ -6,6 +6,9 @@
   * Copyright 1997 Bernd Schmidt
   */
 
+#ifndef UAE_OD_FS_TARGET_H
+#define UAE_OD_FS_TARGET_H
+
 #define TARGET_NAME "fsuae"
 
 #define TARGET_ROM_PATH         "~/"
@@ -31,6 +34,8 @@ int filesys_get_max_frame_packages();
 //void frame_wait_for_filesys();
 //void filesys_handle_events();
 
+void uae_mousehack_helper(int x, int y);
+
 extern int g_uae_deterministic_mode;
 
 static inline int uae_deterministic_mode() {
@@ -39,3 +44,5 @@ static inline int uae_deterministic_mode() {
 
 //void uae_line_update(int line, int updated);
 uint8_t *uae_get_render_buffer();
+
+#endif // UAE_OD_FS_TARGET_H
