@@ -6,18 +6,17 @@ int uae_start_thread_fast (void *(*f)(void *), void *arg,
         uae_thread_id *thread) {
     // FIXME: what is the supposed difference between uae_start_thread and
     // uae_start_thread_fast?
-    return uae_start_thread("(fast)", f, arg, thread);
+    return uae_start_thread(NULL, f, arg, thread);
 }
 
 int uae_start_thread (const char *name, uae_thread_function fn, void *arg,
         uae_thread_id *tid) {
     int result = 1;
-    write_log("uae_start_tread \"%s\" function at %p arg %p\n", name,
-            fn, arg);
-#if 0
-    uae_thread_id thread_id = SDL_CreateThread(fn, arg);
-#endif
-    uae_thread_id thread_id = fs_thread_create(fn, arg);
+    if (name != NULL) {
+        write_log("uae_start_tread \"%s\" function at %p arg %p\n", name,
+                fn, arg);
+    }
+    uae_thread_id thread_id = fs_thread_create(name, fn, arg);
     if (thread_id == NULL) {
         write_log("ERROR creating thread\n");
         result = 0;
@@ -25,6 +24,9 @@ int uae_start_thread (const char *name, uae_thread_function fn, void *arg,
     if (tid) {
         *tid = thread_id;
     }
+    else if (thread_id) {
+        fs_thread_free(thread_id);
+    }
     return result;
 }
 
@@ -35,10 +37,8 @@ int uae_wait_thread (uae_thread_id thread) {
 }
 
 void uae_end_thread (uae_thread_id *thread) {
-    STUB("");
-    //SDL_KillThread(SDL_Thread *thread);
-    fs_thread_destroy(*thread);
     if (thread) {
+        fs_thread_free(*thread);
         *thread = NULL;
     }
 }
